# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

import subprocess
import importlib.util
from shutil import rmtree
from platform import system


def create_file(file, contents=None):
    create_directory(file.parent, empty=False)

    contents = "" if contents is None else contents
    with file.open("w") as file_handle:
        file_handle.write(contents)

    return file


def read_file(file):
    with file.open() as file_handle:
        return file_handle.read()


def delete(path):
    if path.exists():
        if path.is_dir():
            rmtree(path)
        else:
            path.unlink()
    return path


def create_directory(directory, empty=True):
    if empty:
        delete(directory)
    elif directory.exists():
        return directory

    directory.mkdir(parents=True)
    return directory


def run_command(cmd, cwd=None):
    if not isinstance(cmd, list):
        raise ValueError("Must be called with a list, not a string")

    subprocess.check_call(cmd, cwd=cwd)


def load_python_module(file):
    python_module_name = file.stem

    spec = importlib.util.spec_from_file_location(python_module_name, file)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    return module


def system_is_windows():
    return system() == "Windows"



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from pathlib import Path

from tsfpga.hdl_file import HdlFile


def test_file_type():
    assert HdlFile(Path("file.vhd")).is_vhdl
    assert not HdlFile(Path("file.vhd")).is_verilog_source
    assert not HdlFile(Path("file.vhd")).is_verilog_header

    assert HdlFile(Path("file.vh")).is_verilog_header
    assert HdlFile(Path("file.v")).is_verilog_source


def test_can_cast_to_string_without_error():
    str(HdlFile(Path("file.vhd")))



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from pathlib import Path

from tsfpga.build_step_tcl_hook import BuildStepTclHook


def test_step_is_synth():
    assert BuildStepTclHook(Path(), "STEPS.SYNTH_DESIGN.TCL.PRE").step_is_synth
    assert not BuildStepTclHook(Path(), "STEPS.ROUTE_DESIGN.TCL.PRE").step_is_synth


def test_can_cast_to_string_without_error():
    str(BuildStepTclHook(Path("some_file.tcl"), "STEPS.SYNTH_DESIGN.TCL.PRE"))



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from pathlib import Path

import pytest

from tsfpga.constraint import Constraint
from tsfpga.hdl_file import HdlFile


def test_constraint():
    constraint = Constraint(Path("dummy.tcl"))
    constraint.validate_scoped_entity([])

    assert constraint.ref is None
    assert constraint.used_in == "all"

    constraint = Constraint(Path("dummy.tcl"), used_in="impl")
    assert constraint.used_in == "impl"


def test_scoped_constraint(tmp_path):
    constraint = Constraint(tmp_path / "a" / "scoped_constraints" / "apa.tcl", scoped_constraint=True)
    assert constraint.ref == "apa"

    source_files = [HdlFile(tmp_path / "a" / "apa.vhd")]
    constraint.validate_scoped_entity(source_files)


def test_matching_entity_not_existing_should_raise_exception(tmp_path):
    constraint = Constraint(tmp_path / "a" / "scoped_constraints" / "dummy.tcl", scoped_constraint=True)
    with pytest.raises(FileNotFoundError) as exception_info:
        constraint.validate_scoped_entity([])
    assert str(exception_info.value).startswith("Could not find a matching entity file")


def test_can_cast_to_string_without_error():
    str(Constraint(Path("dummy.tcl"), used_in="impl"))



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

import shutil

from tsfpga import TSFPGA_TCL
from tsfpga.system_utils import create_file, read_file
from tsfpga.build_step_tcl_hook import BuildStepTclHook
from .common import run_vivado_tcl, run_vivado_gui
from .size_checker import UtilizationParser
from .tcl import VivadoTcl


class VivadoProject:
    """
    Used for handling a Xilinx Vivado HDL project
    """

    # pylint: disable=too-many-arguments,too-many-instance-attributes
    def __init__(
            self,
            name,
            modules,
            part,
            top=None,
            generics=None,
            constraints=None,
            tcl_sources=None,
            build_step_hooks=None,
            vivado_path=None,
            default_run_index=1,
            defined_at=None,
    ):
        """
        Args:
            name (str): Project name.
            modules (list(:class:`Module <.BaseModule>`)): Modules that shall be included in the project.
            part (str): Part identification.
            top (str): Name of top level entity. If left out, the top level name will be
                inferred from the ``name``.
            generics: A dict with generics values (`dict(name: value)`). Use this parameter
                for "static" generics that do not change between multiple builds of this
                project. These will be set in the project when it is created.

                Compare to the build-time generic argument in :meth:`build`.
            constraints (list(Constraint)): Constraints that will be applied to the project.
            tcl_sources (list(`pathlib.Path`)): A list of TCL files. Use for e.g. block design,
                pinning, settings, etc.
            build_step_hooks (list(BuildStepTclHook)): Build step hooks that will be applied to the project.
            vivado_path (`pathlib.Path`): A path to the Vivado executable. If omitted,
                the default location from the system PATH will be used.
            default_run_index (int): Default run index (synth_X and impl_X) that is set in the project.
                Can also use the argumment to :meth:`build() <VivadoProject.build>` to specify at build-time.
            defined_at (`pathlib.Path`): Optional path to the file where you defined your
                project. To get a useful ``build.py --list`` message. Is useful when you have many
                projects set up.
        """
        self.name = name
        self.modules = modules
        self.part = part
        self.static_generics = generics
        self.constraints = constraints
        self.default_run_index = default_run_index
        self.defined_at = defined_at

        # Will be set by child class when applicable
        self.is_netlist_build = False
        self.analyze_clock_interaction = True

        self.top = name + "_top" if top is None else top
        self._vivado_path = vivado_path

        self._setup_tcl_sources_list(tcl_sources)
        self._setup_build_step_hooks(build_step_hooks)

        self.tcl = VivadoTcl(name=self.name)

    def _setup_tcl_sources_list(self, tcl_sources_from_user):
        # Lists are immutable. Since we assign and modify this one we have to copy it.
        self.tcl_sources = [] if tcl_sources_from_user is None else tcl_sources_from_user.copy()
        self.tcl_sources += [
            TSFPGA_TCL / "vivado_default_run.tcl",
            TSFPGA_TCL / "vivado_fast_run.tcl",
            TSFPGA_TCL / "vivado_messages.tcl",
        ]

    def _setup_build_step_hooks(self, build_step_hooks_from_user):
        # Lists are immutable. Since we assign and modify this one we have to copy it.
        self.build_step_hooks = [] if build_step_hooks_from_user is None else build_step_hooks_from_user.copy()

        self.build_step_hooks.append(BuildStepTclHook(
            TSFPGA_TCL / "vivado_report_utilization.tcl", "STEPS.SYNTH_DESIGN.TCL.POST"))
        self.build_step_hooks.append(BuildStepTclHook(
            TSFPGA_TCL / "vivado_report_utilization.tcl", "STEPS.WRITE_BITSTREAM.TCL.PRE"))
        self.build_step_hooks.append(BuildStepTclHook(
            TSFPGA_TCL / "vivado_check_timing.tcl", "STEPS.WRITE_BITSTREAM.TCL.PRE"))

    def project_file(self, project_path):
        """
        Args:
            project_path (`pathlib.Path`): A path containing a Vivado project.
        Return:
            `pathlib.Path`: The project file of this project, in the given folder
        """
        return project_path / (self.name + ".xpr")

    def _create_tcl(self, project_path, ip_cache_path):
        """
        Make a TCL file that creates a Vivado project
        """
        if project_path.exists():
            raise ValueError(f"Folder already exists: {project_path}")
        project_path.mkdir(parents=True)

        create_vivado_project_tcl = project_path / "create_vivado_project.tcl"
        tcl = self.tcl.create(
            project_folder=project_path,
            modules=self.modules,
            part=self.part,
            top=self.top,
            run_index=self.default_run_index,
            generics=self.static_generics,
            constraints=self.constraints,
            tcl_sources=self.tcl_sources,
            build_step_hooks=self.build_step_hooks,
            ip_cache_path=ip_cache_path,
            disable_io_buffers=self.is_netlist_build
        )
        create_file(create_vivado_project_tcl, tcl)

        return create_vivado_project_tcl

    def create(self, project_path, ip_cache_path=None):
        """
        Create a Vivado project

        Args:
            project_path (`pathlib.Path`): Path where the project shall be placed.
            ip_cache_path (`pathlib.Path`): Path to a folder where the Vivado IP cache can be placed.
                If omitted, the Vivado IP cache mechanism will not be enabled.
        """
        print(f"Creating Vivado project in {project_path}")
        create_vivado_project_tcl = self._create_tcl(project_path, ip_cache_path)
        run_vivado_tcl(self._vivado_path, create_vivado_project_tcl)

    def _build_tcl(self, project_path, output_path, num_threads, run_index, generics, synth_only):
        """
        Make a TCL file that builds a Vivado project
        """
        project_file = self.project_file(project_path)
        if not project_file.exists():
            raise ValueError(f"Project file does not exist in the specified location: {project_file}")

        if self.static_generics is None:
            all_generics = generics
        elif generics is None:
            all_generics = self.static_generics
        else:
            # Add the two dictionaries
            all_generics = self.static_generics.copy()
            all_generics.update(generics)

        build_vivado_project_tcl = project_path / "build_vivado_project.tcl"
        tcl = self.tcl.build(
            project_file=project_file,
            output_path=output_path,
            num_threads=num_threads,
            run_index=run_index,
            generics=all_generics,
            synth_only=synth_only,
            analyze_clock_interaction=self.analyze_clock_interaction
        )
        create_file(build_vivado_project_tcl, tcl)

        return build_vivado_project_tcl

    def pre_build(self, **kwargs):  # pylint: disable=no-self-use, unused-argument
        """
        Override this function in a child class if you wish to do something useful with it.
        Will be called from :meth:`.build` right before the call to Vivado.

        Shall return ``True`` upon success and ``False`` upon failure.

        Args:
            kwargs: Will have all the :meth:`.build` parameters in it. Including additional parameters
                from the user.
        """
        return True

    def post_build(self, **kwargs):  # pylint: disable=no-self-use, unused-argument
        """
        Override this function in a child class if you wish to do something useful with it.
        Will be called from :meth:`.build` right after the call to Vivado.

        Shall return ``True`` upon success and ``False`` upon failure.

        Args:
            kwargs: Will have all the :meth:`.build` parameters in it. Including additional parameters
                from the user. Will also include ``build_result`` with implemented/synthesized size,
                which can be used for asserting the expected resource utilization.
        """
        return True

    def build(self,
              project_path,
              output_path=None,
              run_index=None,
              generics=None,
              synth_only=False,
              num_threads=12,
              **pre_and_post_build_parameters):
        """
        Build a Vivado project

        Args:
            project_path (`pathlib.Path`): A path containing a Vivado project.
            output_path (`pathlib.Path`): Results (bit file, ...) will be placed here.
            run_index (int): Select Vivado run (synth_X and impl_X) to build with.
            generics: A dict with generics values (`dict(name: value)`). Use for run-time
                generics, i.e. values that can change between each build of this project.

                Compare to the create-time generics argument in :meth:`.__init__`.
            synth_only (bool): Run synthesis and then stop.
            num_threads (int): Number of parallell threads to use during run.
            pre_and_post_build_parameters: Additional parameters that will be
                sent to pre- and post build functions.

                .. note::
                    This is a "kwargs" style argument. You can pass any number of named arguments.

        Return:
            BuildResult: Result object with build information.
        """
        synth_only = synth_only or self.is_netlist_build

        if output_path is None and not synth_only:
            raise ValueError("Must specify output_path when doing an implementation run")

        if synth_only:
            print(f"Synthesizing Vivado project in {project_path}")
        else:
            print(f"Building Vivado project in {project_path}, placing artifacts in {output_path}")

        # Run index is optional to specify at build-time
        run_index = self.default_run_index if run_index is None else run_index

        # Make sure register packages are up to date
        for module in self.modules:
            module.create_regs_vhdl_package()

        # Send all available information to pre- and post build functions
        pre_and_post_build_parameters.update(
            project_path=project_path,
            output_path=output_path,
            run_index=run_index,
            generics=generics,
            synth_only=synth_only,
            num_threads=num_threads
        )

        result = BuildResult(self.name)
        result.success = result.success and self.pre_build(**pre_and_post_build_parameters)

        build_vivado_project_tcl = self._build_tcl(project_path=project_path,
                                                   output_path=output_path,
                                                   num_threads=num_threads,
                                                   run_index=run_index,
                                                   generics=generics,
                                                   synth_only=synth_only)
        run_vivado_tcl(self._vivado_path, build_vivado_project_tcl)

        result.synthesis_size = self._get_size(project_path, f"synth_{run_index}")

        if not synth_only:
            impl_folder = project_path / f"{self.name}.runs" / f"impl_{run_index}"
            shutil.copy2(impl_folder / f"{self.top}.bit", output_path / f"{self.name}.bit")
            shutil.copy2(impl_folder / f"{self.top}.bin", output_path / f"{self.name}.bin")
            result.implementation_size = self._get_size(project_path, f"impl_{run_index}")

        # Send result as well to the post-build function
        pre_and_post_build_parameters.update(build_result=result)

        result.success = result.success and self.post_build(**pre_and_post_build_parameters)

        return result

    def open(self, project_path):
        """
        Open the project in Vivado GUI.

        Args:
            project_path (`pathlib.Path`): A path containing a Vivado project.
        """
        run_vivado_gui(self._vivado_path, self.project_file(project_path))

    def _get_size(self, project_path, run):
        """
        Reads the hierarchical utilization report and returns the top level size
        for the specified run.
        """
        report_as_string = read_file(
            project_path / f"{self.name}.runs" / run / "hierarchical_utilization.rpt")
        return UtilizationParser.get_size(report_as_string)

    def __str__(self):
        result = str(self.__class__.__name__)
        if self.defined_at is not None:
            result += f" defined at: {self.defined_at.resolve()}"
        result += "\nName:      " + self.name
        result += "\nTop level: " + self.top
        if self.static_generics is None:
            generics = "-"
        else:
            generics = ", ".join([f"{name}={value}" for name, value in self.static_generics.items()])
        result += "\nGenerics:  " + generics

        return result


class VivadoNetlistProject(VivadoProject):
    """
    Used for handling Vivado build of a module without top level pinning.
    """

    def __init__(
            self,
            analyze_clock_interaction=False,
            result_size_checkers=None,
            **kwargs):
        """
        Args:
            analyze_clock_interaction (bool): Analysis of clock interaction, i.e. checking
                for unhandled clock crossings, is disabled by default. Enabling it will add
                significant build time (can be as much as +40%). Also, in order for clock
                interction check to work, the clocks have to be created using a constraint
                file.
            result_size_checkers (list(:class:`.SizeChecker`)): Size checkers that will be
                executed after a succesful build. Is used to automatically check that
                resource utilization is not greater than expected.
            kwargs: Other arguments accepted by :meth:`.VivadoProject.__init__`.
        """
        super().__init__(**kwargs)

        self.is_netlist_build = True
        self.analyze_clock_interaction = analyze_clock_interaction
        self.result_size_checkers = [] if result_size_checkers is None else result_size_checkers

    def build(self, **kwargs):  # pylint: disable=arguments-differ
        """
        Build the project.

        Args:
            kwargs: All arguments as accepted by :meth:`.VivadoProject.build`.
        """
        result = super().build(**kwargs)
        result.success = result.success and self._check_size(result)

        return result

    def _check_size(self, build_result):
        if not build_result.success:
            print(f"Can not do post_build size check for {self.name} since it did not succeed.")
            return False

        success = True
        for result_size_checker in self.result_size_checkers:
            result = result_size_checker.check(build_result.synthesis_size)
            success = success and result

        return success


class BuildResult:

    """
    Attributes:
        project_name (`str`): The name of the build.
        success (`bool`): True if the build and all pre- and post hooks succeeded.
        synthesis_size (`dict`): A dictionary with the utilization of primitives for the
            synthesized design. Will be ``None`` if synthesis failed or did not run.
        implementation_size (`dict`): A dictionary with the utilization of primitives for
            the implemented design. Will be ``None`` if implementation failed or did not run.
    """

    def __init__(self, name):
        """
        Arguments:
            name (`str`): The name of the build.
        """
        self.name = name
        self.success = True
        self.synthesis_size = None
        self.implementation_size = None



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

import unittest

from tsfpga.markdown_to_html_translator import MarkdownToHtmlTranslator


class TestMarkdownToHtmlTranslator(unittest.TestCase):

    def setUp(self):
        self.markdown_to_html = MarkdownToHtmlTranslator()

    def test_markdown_parser_can_handle_annotating_sentences(self):
        expected = "This sentence <b>should have a large portion</b> in bold face"
        text = r"This sentence **should have a large portion** in bold face"
        assert expected in self.markdown_to_html.translate(text)
        text = r"This sentence __should have a large portion__ in bold face"
        assert expected in self.markdown_to_html.translate(text)

        expected = "This sentence <em>should have a large portion</em> in italics"
        text = "This sentence *should have a large portion* in italics"
        assert expected in self.markdown_to_html.translate(text)
        text = "This sentence _should have a large portion_ in italics"
        assert expected in self.markdown_to_html.translate(text)

    def test_markdown_parser_can_handle_escaped_underscores_and_asterisks(self):
        expected = "Part of this sentence **should be surrounded** by double asterisks"
        text = r"Part of this sentence \*\*should be surrounded\*\* by double asterisks"
        assert expected in self.markdown_to_html.translate(text)
        expected = "Part of this sentence __should be surrounded__ by double underscores"
        text = r"Part of this sentence \_\_should be surrounded\_\_ by double underscores"
        assert expected in self.markdown_to_html.translate(text)

        expected = "Part of this sentence *should be surrounded* by asterisks"
        text = r"Part of this sentence \*should be surrounded\* by asterisks"
        assert expected in self.markdown_to_html.translate(text)
        expected = "Part of this sentence _should be surrounded_ by underscores"
        text = r"Part of this sentence \_should be surrounded\_ by underscores"
        assert expected in self.markdown_to_html.translate(text)

        expected = "Part of this sentence <em>*should be in italics and surrounded*</em> by asterisks"
        text = r"Part of this sentence *\*should be in italics and surrounded\** by asterisks"
        assert expected in self.markdown_to_html.translate(text)
        expected = "Part of this sentence <em>_should be in italics and surrounded_</em> by underscores"
        text = r"Part of this sentence _\_should be in italics and surrounded\__ by underscores"
        assert expected in self.markdown_to_html.translate(text)

        expected = "Part of this sentence *<em>should be in italics and surrounded</em>* by asterisks"
        text = r"Part of this sentence \**should be in italics and surrounded*\* by asterisks"
        assert expected in self.markdown_to_html.translate(text)
        expected = "Part of this sentence _<em>should be in italics and surrounded</em>_ by underscores"
        text = r"Part of this sentence \__should be in italics and surrounded_\_ by underscores"
        assert expected in self.markdown_to_html.translate(text)

        expected = "Part of this sentence should have an <em>*</em> in italics"
        text = r"Part of this sentence should have an *\** in italics"
        assert expected in self.markdown_to_html.translate(text)
        expected = "Part of this sentence should have an <em>_</em> in italics"
        text = r"Part of this sentence should have an _\__ in italics"
        assert expected in self.markdown_to_html.translate(text)



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------


class Constraint:
    """
    Class for handling a constraint file.

    Can handle the regular global constraint files as well as scoped constraints.
    For the latter to work the constraint file name must be the same as the .vhd file name,
    which must be the same as the entity name.
    """

    def __init__(self, file, used_in="all", scoped_constraint=False, processing_order="normal"):
        """
        Args:
            file (`pathlib.Path`): Path to the constraint file. Typically ends in .xdc or .tcl.
            used_in (str): Optionally the constraint can be enabled only for "synth" or "impl".
            scoped_constraint (bool): If enabled the constraint file will be loaded with the "-ref"
                argument in Vivado. An entity with the same name must exist.
            processing_order (str): Optionally the processing order can be changed to "early" or "late".
        """
        self.file = file
        self.used_in = used_in
        self.ref = file.stem if scoped_constraint else None
        self.processing_order = processing_order.lower()

        assert self.used_in in ["all", "synth", "impl"], self.used_in
        assert self.processing_order in ["early", "normal", "late"], self.processing_order

    def validate_scoped_entity(self, source_files):
        """
        Make sure that a matching entity file exists in case this is a scoped constraint.
        The list of source files should be the synthesis files for the module that this
        constraint belongs to.
        """
        if self.ref is not None:
            if not any([source_file.path.stem == self.ref] for source_file in source_files):
                raise FileNotFoundError(f"Could not find a matching entity file for scoped constraint file {self.file}")
        return True

    def __str__(self):
        return str(self.file)



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from unittest import mock
import io

from tsfpga.vivado.size_checker import UtilizationParser, EqualTo, LessThan, TotalLuts


def test_utilzation_parser():
    report = " \
Writing top or top) or (top here should not cause issues\n\
Also writing pipes | | | | should be fine\n\
Random sections of dashes should be fine:\n\
---------------------------------------------------------------\n\
-----------------------+----------------------------+----------\n\
But this should hit:\n\
| x | x | a | b | c | d |\n\
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\
| x | (top) | 1 | 2 | 3 | 4 |\n\
"

    result = UtilizationParser.get_size(report)
    assert result["a"] == 1
    assert result["b"] == 2
    assert result["c"] == 3
    assert result["d"] == 4
    assert len(result) == 4


@mock.patch("sys.stdout", new_callable=io.StringIO)
def test_size_checker_less_than_pass(mock_stdout):
    result_size = {"Total LUTs": 5}
    assert TotalLuts(LessThan(10)).check(result_size)
    assert mock_stdout.getvalue() == "Result size check passed for Total LUTs: 5 (< 10)\n"


@mock.patch("sys.stdout", new_callable=io.StringIO)
def test_size_checker_equal_to_fail(mock_stdout):
    result_size = {"Total LUTs": 5}
    assert not TotalLuts(EqualTo(50)).check(result_size)
    assert mock_stdout.getvalue() == "Result size check failed for Total LUTs. Got 5, expected 50.\n"



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from pathlib import Path
from unittest import mock, TestCase

import pytest

from tsfpga.module import BaseModule, get_modules
from tsfpga.system_utils import create_file, create_directory


def test_file_list_filtering(tmp_path):
    module_name = "zebra"
    path = tmp_path / module_name

    create_directory(path / "folder_should_not_be_included")
    create_file(path / "should_not_be_included.apa")

    synth_files = [create_file(path / "syn.v"),
                   create_file(path / "rtl" / "syn.v"),
                   create_file(path / "src" / "syn.vhd"),
                   create_file(path / "hdl" / "rtl" / "syn.vhd"),
                   create_file(path / "hdl" / "package" / "syn.vhd")]

    test_files = [create_file(path / "test" / "test.v"),
                  create_file(path / "rtl" / "tb" / "test.vhd")]

    sim_files = [create_file(path / "sim" / "sim.vhd")]

    my_module = BaseModule(path, "zebra")

    files = [file.path for file in my_module.get_synthesis_files()]
    assert set(files) == set(synth_files)

    files = [file.path for file in my_module.get_simulation_files()]
    assert set(files) == set(synth_files + test_files + sim_files)

    files = [file.path for file in my_module.get_simulation_files(include_tests=False)]
    assert set(files) == set(synth_files + sim_files)


def test_scoped_constraints(tmp_path):
    module_path = tmp_path / "apa"
    create_file(module_path / "src" / "hest.vhd")
    create_file(module_path / "scoped_constraints" / "hest.tcl")

    my_module = BaseModule(module_path, "apa")
    scoped_constraints = my_module.get_scoped_constraints()
    assert len(scoped_constraints) == 1
    assert scoped_constraints[0].ref == "hest"


def test_scoped_constraint_entity_not_existing_should_raise_error(tmp_path):
    module_path = tmp_path / "apa"
    create_file(module_path / "scoped_constraints" / "hest.tcl")

    module = BaseModule(module_path, "apa")
    with pytest.raises(FileNotFoundError) as exception_info:
        module.get_scoped_constraints()
    assert str(exception_info.value).startswith("Could not find a matching entity file")


def test_can_cast_to_string_without_error():
    str(BaseModule(Path("dummy"), "dummy"))


def test_test_case_name():
    assert BaseModule.test_case_name(
        generics=dict(apa=3, hest_zebra="foo")) == "apa_3.hest_zebra_foo"
    assert BaseModule.test_case_name(
        name="foo", generics=dict(apa=3, hest_zebra="bar")) == "foo.apa_3.hest_zebra_bar"


@pytest.mark.usefixtures("fixture_tmp_path")
class TestGetModules(TestCase):

    tmp_path = None

    def setUp(self):
        create_directory(self.tmp_path / "a")
        create_directory(self.tmp_path / "b")
        create_directory(self.tmp_path / "c")

        self.modules_folders = [self.tmp_path]

    def test_name_filtering_include(self):
        modules = get_modules(self.modules_folders, names_include=["a", "b"])
        assert set(module.name for module in modules) == set(["a", "b"])

    def test_name_filtering_avoid(self):
        modules = get_modules(self.modules_folders, names_avoid=["a", "b"])
        assert set(module.name for module in modules) == set(["c"])

    def test_name_filtering_include_and_avoid(self):
        modules = get_modules(self.modules_folders,
                              names_include=["a", "c"],
                              names_avoid=["b", "c"])
        assert set(module.name for module in modules) == set(["a"])

    def test_library_name_does_not_have_lib_suffix(self):
        modules = get_modules(self.modules_folders)
        assert set(module.library_name for module in modules) == set(["a", "b", "c"])

    def test_library_name_has_lib_suffix(self):
        modules = get_modules(self.modules_folders, library_name_has_lib_suffix=True)
        assert set(module.library_name for module in modules) == set(["a_lib", "b_lib", "c_lib"])

    def test_stray_file_can_exist_in_modules_folder_without_error(self):
        create_file(self.tmp_path / "text_file.txt")
        modules = get_modules(self.modules_folders)
        assert len(modules) == 3

    def test_local_override_of_module_type(self):
        module_file_content = """
from tsfpga.module import BaseModule

class Module(BaseModule):

    def id(self):
        return """

        create_file(self.tmp_path / "a" / "module_a.py", module_file_content + "\"a\"")
        create_file(self.tmp_path / "b" / "module_b.py", module_file_content + "\"b\"")

        modules = get_modules(self.modules_folders)

        assert len(modules) == 3
        for module in modules:
            if module.name == "a":
                assert module.id() == "a"
            elif module.name == "b":
                assert module.id() == "b"
            elif module.name == "c":
                assert isinstance(module, BaseModule)
            else:
                assert False

    @mock.patch("tsfpga.module.from_toml", autospec=True)
    def test_register_object_creation_synthesis(self, from_toml):
        toml_file = create_file(self.tmp_path / "a" / "regs_a.toml")

        module = get_modules(self.modules_folders, names_include=["a"])[0]
        module.get_synthesis_files()
        module.get_synthesis_files()

        from_toml.assert_called_once_with("a", toml_file, mock.ANY)

    @mock.patch("tsfpga.module.from_toml", autospec=True)
    def test_register_object_creation_simulation(self, from_toml):
        toml_file = create_file(self.tmp_path / "a" / "regs_a.toml")

        module = get_modules(self.modules_folders, names_include=["a"])[0]
        module.get_simulation_files()
        module.get_simulation_files()

        from_toml.assert_called_once_with("a", toml_file, mock.ANY)

    @mock.patch("tsfpga.module.from_toml", autospec=True)
    def test_register_object_creation_mixed(self, from_toml):
        toml_file = create_file(self.tmp_path / "a" / "regs_a.toml")

        module = get_modules(self.modules_folders, names_include=["a"])[0]
        module.get_synthesis_files()
        module.get_simulation_files()

        from_toml.assert_called_once_with("a", toml_file, mock.ANY)

    def test_deprecated_register_json_file_should_raise_exception(self):
        json_file = create_file(self.tmp_path / "a" / "regs_a.json")

        module = get_modules(self.modules_folders, names_include=["a"])[0]
        with pytest.raises(ValueError) as exception_info:
            module.get_synthesis_files()
        assert str(exception_info.value).startswith(
            f"DEPRECATED: Using deprecated json file name: {json_file}")



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

import argparse
from pathlib import Path
import re
import sys

PATH_TO_TSFPGA = Path(__file__).parent.parent
sys.path.append(str(PATH_TO_TSFPGA.resolve()))

from tsfpga.git_utils import find_git_files
from tsfpga.system_utils import create_file, read_file
from tsfpga.test.lint.test_copyright import CopyrightHeader, files_to_check_for_copyright_header


RE_TRAILING_WHITESPACE = re.compile(" +\n", re.DOTALL)
RE_TAB = re.compile("\t", re.DOTALL)


def fix_trailing_whitespace(file):
    contents = read_file(file)

    if RE_TRAILING_WHITESPACE.search(contents):
        print(f"Fixing trailing whitespace in {file}")
        create_file(file, RE_TRAILING_WHITESPACE.sub("\n", contents))


def fix_tabs(file, tab_width):
    contents = read_file(file)

    if RE_TAB.search(contents):
        print(f"Fixing tabs in {file}")
        replacement = " " * tab_width
        create_file(file, RE_TAB.sub(replacement, contents))


def fix_lint(files, tab_width):
    for file in files:
        fix_trailing_whitespace(file)
        fix_tabs(file, tab_width)


def arguments():
    parser = argparse.ArgumentParser("Automatically fix some lint problems")
    parser.add_argument("--tab-width", type=int, default=2, help="number of spaces to use when replacing tab characters")
    parser.add_argument("files", nargs="*", help="fixup these specific files")
    return parser.parse_args()


def main():
    args = arguments()

    files = args.files if args.files else list(find_git_files(file_endings_avoid=(".png")))
    fix_lint(files, args.tab_width)

    files = args.files if args.files else list(files_to_check_for_copyright_header())
    for filename in files:
        copyright_header = CopyrightHeader("Lukas Vik", filename)
        if not copyright_header.check_file():
            copyright_header.fix_file()


if __name__ == "__main__":
    main()



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from tsfpga.module import BaseModule


class Module(BaseModule):

    def setup_simulations(self, vunit_proj, **kwargs):
        tb = vunit_proj.library(self.library_name).test_bench("tb_axil_reg_file")
        tb.test("read_from_non_existent_register").set_generic("use_axil_bfm", False)
        tb.test("read_from_non_read_type_register").set_generic("use_axil_bfm", False)
        tb.test("write_to_non_existent_register").set_generic("use_axil_bfm", False)
        tb.test("write_to_non_write_type_register").set_generic("use_axil_bfm", False)



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

import argparse
from pathlib import Path
import json
from shutil import copy2, make_archive
import sys


PATH_TO_TSFPGA = Path(__file__).parent.parent
sys.path.append(str(PATH_TO_TSFPGA))
PATH_TO_VUNIT = PATH_TO_TSFPGA.parent / "vunit"
sys.path.append(str(PATH_TO_VUNIT))

from vunit.color_printer import COLOR_PRINTER, NO_COLOR_PRINTER

from tsfpga.build_project_list import BuildProjectList
from tsfpga.system_utils import create_directory, delete
from tsfpga.vivado.project import BuildResult

from examples.tsfpga_example_env import get_tsfpga_modules, TSFPGA_EXAMPLES_TEMP_DIR


def arguments(default_temp_dir=TSFPGA_EXAMPLES_TEMP_DIR):
    parser = argparse.ArgumentParser("Create, synth and build an FPGA project",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--no-color",
                        action="store_true",
                        help="disable color in printouts")
    parser.add_argument("--list-only",
                        action="store_true",
                        help="list the available projects")
    parser.add_argument("--open",
                        action="store_true",
                        help="open an existing project in the GUI")
    parser.add_argument("--use-existing-project",
                        action="store_true",
                        help="build an existing project (do not create)")
    parser.add_argument("--generate-registers-only",
                        action="store_true",
                        help="only generate the register artifacts (C/C++ code, HTML, ...) for inspection")
    parser.add_argument("--create-only",
                        action="store_true",
                        help="only create a project")
    parser.add_argument("--synth-only",
                        action="store_true",
                        help="only synthesize a project")
    parser.add_argument("--netlist-builds",
                        action="store_true",
                        help="list netlist build projects instead of top level build projects")
    parser.add_argument("--project-path",
                        type=Path,
                        default=default_temp_dir / "projects",
                        help="the FPGA build project will be placed here")
    parser.add_argument("--ip-cache-path",
                        type=Path,
                        default=default_temp_dir / "vivado_ip_cache",
                        help="location of Vivado IP cache")
    parser.add_argument("--output-path",
                        type=Path,
                        required=False,
                        help="the output products (bit file, ...) will be placed here")
    parser.add_argument("--num-threads",
                        type=int,
                        default=8,
                        help="number of threads to use when building project")
    parser.add_argument("project_filters",
                        nargs="*",
                        help="filter for which projects to build")
    parser.add_argument("--collect-artifacts-only",
                        action="store_true",
                        help="only create release zip from an existing build result")
    args = parser.parse_args()

    return args


def main():
    modules = get_tsfpga_modules()
    projects = BuildProjectList(modules)
    args = arguments()

    sys.exit(setup_and_run(modules, projects, args))


def setup_and_run(modules, projects, args):
    """
    Returns 0 if everything passed, otherwise non-zero.
    """
    if args.list_only:
        print(projects.list_projects(args.project_filters, args.netlist_builds))
        return 0

    if args.generate_registers_only:
        # Generate register output from all modules. Note that this is not used by the build
        # flow or simulation flow, it is only for the user to inspect the artifacts.
        generate_registers(modules, args.project_path.parent / "registers")
        return 0

    if not args.project_filters:
        print("Must explicitly select builds. Available projects are:")
        print(projects.list_projects(args.project_filters, args.netlist_builds))
        return -1

    build_results = []
    for project in projects.get_projects(args.project_filters, args.netlist_builds):
        project_path = args.project_path / project.name

        if args.open:
            project.open(project_path)
            continue

        output_path = args.output_path if args.output_path else project_path

        # Possible that the user wants to collect artifacts (bit file) from a previous build
        if not args.collect_artifacts_only:
            build_results.append(build(args, project, project_path, output_path))

        if args.create_only or args.synth_only:
            continue

        if not project.is_netlist_build:
            collect_artifacts(project, output_path)

    if build_results:
        return print_results(build_results, args.no_color)
    return 0


def build(args, project, project_path, output_path):
    if not args.use_existing_project:
        project.create(project_path=project_path, ip_cache_path=args.ip_cache_path)

    if args.create_only:
        # Return BuildResult object with success, but no sizes.
        return BuildResult(project.name)

    create_directory(output_path, empty=False)
    result = project.build(project_path=project_path,
                           output_path=output_path,
                           synth_only=args.synth_only,
                           num_threads=args.num_threads)
    return result


def print_results(build_results, no_color):
    """
    Print result summary. Return 0 if all passed otherwise non-zero.
    """
    printer = NO_COLOR_PRINTER if no_color else COLOR_PRINTER
    green = COLOR_PRINTER.GREEN + COLOR_PRINTER.INTENSITY
    red = COLOR_PRINTER.RED + COLOR_PRINTER.INTENSITY

    name_length = max([len(build_result.name) for build_result in build_results])
    separator_length = max(len("pass") + 1 + name_length, 40)
    separator = "=" * separator_length

    num_pass = 0
    num_fail = 0
    num_total = len(build_results)

    print("\n" + ("=" * 4) + " Summary " + ("=" * (len(separator) - 4 - len(" Summary "))))
    for build_result in build_results:
        if build_result.success:
            num_pass += 1
            printer.write("pass", fg=green)
        else:
            num_fail += 1
            printer.write("fail", fg=red)

        print(f" {build_result.name}")

        if build_result.implementation_size:
            build_step = "Implementation"
            size = build_result.implementation_size
        else:
            build_step = "Synthesis"
            size = build_result.synthesis_size

        if size:
            print(f"{build_step} size:")
            print(json.dumps(size, indent=2))
            print("")

    print(separator)
    if num_pass > 0:
        printer.write("pass", fg=green)
        print(f" {num_pass} of {num_total}")
    if num_fail > 0:
        printer.write("fail", fg=red)
        print(f" {num_fail} of {num_total}")

    print(separator)
    if num_fail == 0:
        printer.write("All passed!", fg=green)
        print()
    else:
        printer.write("Some failed!", fg=red)
        print()

    return num_fail


def generate_registers(modules, output_path):
    print(f"Generating registers in {output_path.resolve()}")

    for module in modules:
        # Create the package in the modules' folder as well, for completeness sake
        module.create_regs_vhdl_package()

        if module.registers is not None:
            vhdl_path = create_directory(output_path / "vhdl", empty=False)
            module.registers.create_vhdl_package(vhdl_path)

            module.registers.copy_source_definition(output_path / "toml")

            module.registers.create_c_header(output_path / "c")
            module.registers.create_cpp_interface(output_path / "cpp" / "include")
            module.registers.create_cpp_header(output_path / "cpp" / "include")
            module.registers.create_cpp_implementation(output_path / "cpp")
            module.registers.create_html_page(output_path / "doc")
            module.registers.create_html_table(output_path / "doc" / "tables")


def collect_artifacts(project, output_path):
    version = "0.0.0.0"
    release_dir = create_directory(output_path / f"{project.name}-{version}", empty=True)
    print(f"Creating release in {release_dir.resolve()}.zip")

    generate_registers(project.modules, release_dir / "registers")
    copy2(output_path / f"{project.name }.bit", release_dir)
    copy2(output_path / f"{project.name}.bin", release_dir)
    if (output_path / f"{project.name}.xsa").exists():
        copy2(output_path / f"{project.name}.xsa", release_dir)

    make_archive(release_dir, "zip", release_dir)

    # Remove folder so that only zip remains
    delete(release_dir)


if __name__ == "__main__":
    main()



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

import unittest

import pytest

from tsfpga.build_project_list import BuildProjectList


class TestBuildProjectList(unittest.TestCase):

    def setUp(self):
        modules = [DummyModule("one"), DummyModule("two")]
        modules += [DummyNetlistModule("three"), DummyNetlistModule("four"), DummyNetlistModule("five")]
        self.projects = BuildProjectList(modules)

    def test_get_project(self):
        projects = self.projects.get_projects(["one"], False)
        assert len(projects) == 1
        expected = {"one"}
        assert {project.name for project in projects} == expected

        projects = self.projects.get_projects(["*o*"], False)
        assert len(projects) == 2
        expected = {"one", "two"}
        assert {project.name for project in projects} == expected

        projects = self.projects.get_projects(["*o*"], True)
        assert len(projects) == 1
        expected = {"four"}
        assert {project.name for project in projects} == expected

        projects = self.projects.get_projects(["*e*"], True)
        assert len(projects) == 2
        expected = {"three", "five"}
        assert {project.name for project in projects} == expected

        projects = self.projects.get_projects(["four", "*e*"], True)
        assert len(projects) == 3
        expected = {"three", "four", "five"}
        assert {project.name for project in projects} == expected

    def test_get_project_with_name_that_does_not_exist_shall_raise_exception(self):
        with pytest.raises(ValueError) as exception_info:
            self.projects.get_projects(["f"], False)
        assert str(exception_info.value) == "Could not find projects with filters: ['f']"
        with pytest.raises(ValueError) as exception_info:
            self.projects.get_projects(["one"], True)
        assert str(exception_info.value) == "Could not find projects with filters: ['one']"
        with pytest.raises(ValueError) as exception_info:
            self.projects.get_projects(["three"], False)
        assert str(exception_info.value) == "Could not find projects with filters: ['three']"

    def test_can_cast_to_string_without_error(self):
        str(self.projects)


class DummyModule:

    def __init__(self, name):
        self.name = name

    def get_build_projects(self):
        return [DummyBuildProject(self.name, False)]


class DummyNetlistModule:

    def __init__(self, name):
        self.name = name

    def get_build_projects(self):
        return [DummyBuildProject(self.name, True)]


class DummyBuildProject:

    def __init__(self, name, is_netlist_build):
        self.name = name
        self.is_netlist_build = is_netlist_build



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from tsfpga.system_utils import create_file
from tsfpga.fix_lint import fix_trailing_whitespace, fix_tabs
from tsfpga.test import file_equals


def test_fix_trailing_whitespace(tmp_path):
    data = "Apa \nhest    \nzebra"
    data_fixed = "Apa\nhest\nzebra"

    file = create_file(tmp_path / "temp_file_for_test.txt", data)
    fix_trailing_whitespace(file)
    assert file_equals(file, data_fixed)


def test_fix_tabs(tmp_path):
    data = "Apa\thest \t zebra"
    data_fixed = "Apa hest   zebra"
    file = create_file(tmp_path / "data_width_1.txt", data)
    fix_tabs(file, tab_width=1)
    assert file_equals(file, data_fixed)

    data_fixed = "Apa  hest    zebra"
    file = create_file(tmp_path / "data_width_2.txt", data)
    fix_tabs(file, tab_width=2)
    assert file_equals(file, data_fixed)



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

import re
import subprocess


def get_svn_revision_information(cwd=None):
    check_that_svn_commands_are_available(cwd)
    result = "r" + get_svn_revision(cwd)
    if svn_local_changes_are_present(cwd):
        result += " (local changes present)"
    return result


def svn_commands_are_available(cwd=None):
    try:
        get_svn_revision(cwd)
    except subprocess.CalledProcessError:
        return False
    return True


def check_that_svn_commands_are_available(cwd=None):
    if not svn_commands_are_available(cwd):
        mesg = "Could not run svn. Is the command available on PATH? Is the script called from a repo?"
        raise RuntimeError(mesg)


def get_svn_revision(cwd=None):
    command = ["svn", "info", "--show-item", "revision"]
    output = subprocess.check_output(command, cwd=cwd, universal_newlines=True)
    # Remove trailing newline
    return output.strip()


def svn_local_changes_are_present(cwd=None):
    """
    Return true if the repo contains changes that have been made after the last commit.
    Info from here: https://rubyinrails.com/2014/01/11/svn-command-to-check-modified-files/
    """
    command = ["svn", "status"]
    output = subprocess.check_output(command, cwd=cwd, universal_newlines=True)
    # Status code for file Added, Deleted, Modified, in Conflict or missing
    regexp = re.compile(r"\n[ADMC!] ")
    return regexp.search(output) is not None



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from pathlib import Path
import platform
from shutil import which, make_archive
import zipfile

from tsfpga.system_utils import create_file, delete


class VivadoSimlibCommon:

    """
    Class for handling Vivado simlib used for simulation. Keeps track of when a
    (re)compile is needed.
    """

    _libraries = None
    _vivado_path = None
    _output_path = None

    def compile_if_needed(self):
        """
        Compile if needed (if :meth:`compile_is_needed <.compile_is_needed>` condition is not fulfilled).
        """
        if self.compile_is_needed:
            self.compile()
            return True
        return False

    @property
    def compile_is_needed(self):
        """
        If there is compiled simlib available that matches

         * Operating system
         * Vivado version
         * Simulator version

        then there should not be a recompile.

        .. note::
            Child implementations might add further conditions.

        Return:
            True if compiled simlib is not available. False otherwise.
        """
        if self._done_token.exists():
            return False
        return True

    def compile(self):
        """
        Compile simlib.
        """
        delete(self._done_token)
        print(f"Compiling Vivado simlib in {self._output_path}")

        self._compile()

        create_file(self._done_token, "Done!")

    def _compile(self):
        """
        Compile simlib.
        """
        raise NotImplementedError()

    def add_to_vunit_project(self):
        """
        Add the compiled simlib to your VUnit project.
        """
        self._add_to_vunit_project()

    def _add_to_vunit_project(self):
        raise NotImplementedError()

    @property
    def artifact_name(self):
        """
        str: The name of the folder where simlib is or will be compiled.
        Follows a format ``vivado-simlib-WW.XX.YY.ZZ`` suitable for storage and versioning
        in Artifactory.
        """
        return self._output_path.name

    def to_archive(self):
        """
        Compress compiled simlib to an archive.

        Return:
            `pathlib.Path`: Path to the archive.
        """
        make_archive(self._output_path, "zip", self._output_path)
        archive = self._output_path.parent / (self._output_path.name + ".zip")
        return archive

    def from_archive(self, archive):
        """
        Unpack compiled simlib from an existing archive.

        Args:
            archive (`pathlib.Path`): Path to a zip archive with previously compiled simlib.
        """
        with zipfile.ZipFile(archive, "r") as zip_handle:
            zip_handle.extractall(self._output_path)

    def _get_version_tag(self):
        tag = "vivado-simlib-"
        tag += self._get_operating_system_tag()
        tag += "." + self._get_vivado_version_tag()
        tag += "." + self._get_simulator_tag()
        return tag

    def _get_operating_system_tag(self):
        """
        Return e.g. "linux".
        """
        return self._format_version(platform.system())

    def _get_vivado_version_tag(self):
        """
        Return e.g. "vivado_2019_2".
        """
        vivado_path = self._vivado_path
        if vivado_path == "vivado":
            vivado_path = which(vivado_path)
            assert vivado_path is not None, "Could not find vivado location"
        vivado_version = Path(vivado_path).parent.parent.name
        return self._format_version("vivado_" + vivado_version)

    def _get_simulator_tag(self):
        raise NotImplementedError()

    @staticmethod
    def _format_version(version):
        """
        Format version string to something suitable fort artifactory versioning.
        """
        return version.replace(".", "_").replace("-", "_").lower()

    @property
    def _done_token(self):
        """
        Path to "done" token file.
        """
        return self._output_path / "done.txt"



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

import re


class MarkdownToHtmlTranslator:

    def __init__(self):
        self._compile_markdown_parser()

    def _compile_markdown_parser(self):
        r"""
        Strong: **double asterisks** or __double underscores__
        Emphasis: *single asterisks* or _single underscores_
        Literal asterisks or underscores are escaped: \* \_
        """
        not_escaped = r"(?<!\\)"
        double_asterisks = r"\*\*"
        double_underscores = r"\_\_"
        single_asterisk = r"\*"
        single_underscore = r"\_"
        match_text = r"(.*?)"

        # These patterns match underscores and asterisks only if they are not preceded by \escape
        self.strong_pattern1 = re.compile(
            not_escaped + double_asterisks + match_text + not_escaped + double_asterisks)
        self.strong_pattern2 = re.compile(
            not_escaped + double_underscores + match_text + not_escaped + double_underscores)
        self.em_pattern1 = re.compile(
            not_escaped + single_asterisk + match_text + not_escaped + single_asterisk)
        self.em_pattern2 = re.compile(
            not_escaped + single_underscore + match_text + not_escaped + single_underscore)

        # This pattern matches escaped underscores and asterisks
        self.escaped_literal_pattern = re.compile(r"\\(\*|_)")

    def translate(self, text):
        text = re.sub(self.strong_pattern1, r"<b>\g<1></b>", text)
        text = re.sub(self.strong_pattern2, r"<b>\g<1></b>", text)
        text = re.sub(self.em_pattern1, r"<em>\g<1></em>", text)
        text = re.sub(self.em_pattern2, r"<em>\g<1></em>", text)
        # Remove the escape sign
        text = re.sub(self.escaped_literal_pattern, r"\g<1>", text)
        return text



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

import copy

from tsfpga.constraint import Constraint
from tsfpga.hdl_file import HdlFile
from tsfpga.register_list import from_toml
from tsfpga.system_utils import load_python_module


class BaseModule:
    """
    Base class for handling a HDL module with RTL code, constraints, etc.

    Files are gathered from a lot of different subfolders, to accommodate for projects having
    different catalog structure.
    """

    def __init__(self, path, library_name, default_registers=None):
        """
        Args:
            path (`pathlib.Path`): Path to the module folder.
            library_name (str): VHDL library name.
            default_registers (list(Register)): Default registers.
        """
        self.path = path.resolve()
        self.name = path.name
        self.library_name = library_name

        # Note: Likely mutable object, need to create deep copy before using.
        self._default_registers = default_registers
        self._registers = None

    @staticmethod
    def _get_file_list(folders, file_endings):
        """
        Returns a list of files given a list of folders.
        """
        files = []
        for folder in folders:
            for file in folder.glob("*"):
                if file.is_file() and file.name.lower().endswith(file_endings):
                    files.append(file)
        return files

    def _get_hdl_file_list(self, folders):
        """
        Return a list of HDL file objects.
        """
        return [HdlFile(filename) for filename in self._get_file_list(folders, HdlFile.file_endings)]

    @property
    def registers(self):
        """
        :class:`.RegisterList`: Get the registers for this module.
        """
        if self._registers is not None:
            # Only create object once
            return self._registers

        toml_file = self.path / (f"regs_{self.name}.toml")
        if toml_file.exists():
            self._registers = from_toml(self.name, toml_file, copy.deepcopy(self._default_registers))
            self.registers_hook()
            return self._registers

        for deprecated_json_file in [self.path / f"{self.name}_regs.json", self.path / f"regs_{self.name}.json"]:
            if deprecated_json_file.exists():
                message = f"DEPRECATED: Using deprecated json file name: {deprecated_json_file}"
                message += f"\nDEPRECATED: Use TOML file instead: {toml_file}"
                raise ValueError(message)

        return None

    def registers_hook(self):
        """
        This function will be called directly after creating this module's registers from
        the TOML definition file.

        This is a good place if you want to add or modify some registers from Python.
        Override this method and implement the desired behavior in a child class.

        .. Note::
            This default method does nothing and can be safely overridden.
        """

    def create_regs_vhdl_package(self):
        """
        Create a VHDL package in this module with register definitions.
        """
        if self.registers is not None:
            self.registers.create_vhdl_package(self.path)

    def get_synthesis_files(self):
        """
        Return:
            list(:class:`.HdlFile`): Files that should be included in a synthesis project.
        """
        self.create_regs_vhdl_package()

        folders = [
            self.path,
            self.path / "src",
            self.path / "rtl",
            self.path / "hdl" / "rtl",
            self.path / "hdl" / "package",
        ]
        return self._get_hdl_file_list(folders)

    def get_simulation_files(self, include_tests=True):
        """
        Args:
            include_tests (bool): When False the test folder is not included.
                The use case of include_tests is when testing a primary module
                that depends on other secondary modules, we may want to compile
                the simulation files (``sim`` folder) of the secondary modules but not their
                test files (``test`` folder).

                .. Note::
                    `test` files are considered private to the module and should never be used by
                    other modules.

        Return:
            list(:class:`.HdlFile`): Files that should be included in a simulation project.
        """
        self.create_regs_vhdl_package()

        test_folders = [
            self.path / "sim",
        ]

        if include_tests:
            test_folders += [self.path / "rtl" / "tb",
                             self.path / "test"]

        return self.get_synthesis_files() + self._get_hdl_file_list(test_folders)

    def setup_simulations(self, vunit_proj, **kwargs):
        """
        Setup local configuration of this module's test benches.

        .. Note::
            This default method does nothing. Should be overridden by modules that have
            any test benches that operate via generics.

        Args:
            vunit_proj: The VUnit project that is used to run simulation.
            kwargs: Use this to pass an arbitrary list of arguments from your ``simulate.py``
                to the module where you set up your tests. This could be, e.g., data dimensions,
                location of test files, etc.
        """

    def setup_formal(self, formal_proj, **kwargs):
        """
        Setup this module's formal tests.

        .. Note::
            This default method does nothing. Should be overridden by modules that
            utilize formal verification.
        """

    def get_ip_core_files(self):
        """
        Return:
            list(`pathlib.Path`): TCL files that set up the IP cores from this module.
        """
        folders = [
            self.path / "ip_cores",
        ]
        file_endings = ("tcl")
        return self._get_file_list(folders, file_endings)

    def get_scoped_constraints(self):
        """
        Return:
            list(:class:`.Constraint`): Constraints that will be applied to a certain
            entity within the module.
        """
        scoped_constraints_folders = [
            self.path / "scoped_constraints",
            self.path / "entity_constraints",
            self.path / "hdl" / "constraints",
        ]
        constraints_file_endings = ("tcl", "xdc")
        constraint_files = self._get_file_list(scoped_constraints_folders, constraints_file_endings)

        constraints = []
        if constraint_files:
            synthesis_files = self.get_synthesis_files()
            for constraint_file in constraint_files:
                # Scoped constraints often depend on clocks having been created by another constraint
                # file before they can work. Set processing order to "late" to make this more probable.
                constraint = Constraint(constraint_file, scoped_constraint=True, processing_order="late")
                constraint.validate_scoped_entity(synthesis_files)
                constraints.append(constraint)
        return constraints

    def get_build_projects(self):  # pylint: disable=no-self-use
        """
        Get FPGA build projects defined by this module.

        .. Note::
            Should be overridden by modules that set up build projects.

        Return:
            list(:class:`.VivadoProject`): FPGA build projects.
        """
        return []

    @staticmethod
    def test_case_name(name=None, generics=None):
        """
        Costruct a string suitable for naming test cases.

        Example result: MyName.GenericA_ValueA.GenericB_ValueB
        """
        if name:
            test_case_name = name
        else:
            test_case_name = ""

        if generics:
            generics_string = ".".join([f"{key}_{value}" for key, value in generics.items()])
            if test_case_name:
                test_case_name = f"{name}.{generics_string}"
            else:
                test_case_name = generics_string

        return test_case_name

    def add_config(self, test, name=None, generics=None):
        """
        Add config for VUnit test case. Wrapper that sets a suitable name.
        """
        name = self.test_case_name(name, generics)
        test.add_config(name=name, generics=generics)

    def __str__(self):
        return f"{self.name}:{self.path}"


def iterate_module_folders(modules_folders):
    for modules_folder in modules_folders:
        for module_folder in modules_folder.glob("*"):
            if module_folder.is_dir():
                yield module_folder


def get_module_object(path, name, library_name_has_lib_suffix, default_registers):
    module_file = path / f"module_{name}.py"
    library_name = f"{name}_lib" if library_name_has_lib_suffix else name

    if module_file.exists():
        return load_python_module(module_file).Module(path, library_name, default_registers)
    return BaseModule(path, library_name, default_registers)


def get_modules(modules_folders,
                names_include=None,
                names_avoid=None,
                library_name_has_lib_suffix=False,
                default_registers=None):
    """
    Get a list of Module objects based on the source code folders.

    Args:
        modules_folders (list(`pathlib.Path`)): A list of paths where your modules are located.
        names_include (list(str)): If specified, only modules with these names will be included.
        names_avoid (list(str)): If specified, modules with these names will be discarded.
        library_name_has_lib_suffix (bool): If set, the library name will be
            ``<module name>_lib``, otherwise it is just ``<module name>``.
        default_registers (list(Register)): Default registers.

    Return:
        List of module objects (:class:`BaseModule` or child classes thereof) created from
        the specified folders.
    """
    modules = []

    for module_folder in iterate_module_folders(modules_folders):
        module_name = module_folder.name
        if (names_include is None or module_name in names_include) \
                and (names_avoid is None or module_name not in names_avoid):
            modules.append(get_module_object(module_folder,
                                             module_name,
                                             library_name_has_lib_suffix,
                                             default_registers))

            deprecated_project_file = module_folder / f"project_{module_name}.py"
            if deprecated_project_file.exists():
                message = f"DEPRECATED: Using deprecated project file: {deprecated_project_file}"
                message += f"\nDEPRECATED: Instead use module_{module_name}.py"
                raise ValueError(message)

    return modules



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from tsfpga.module import BaseModule


class Module(BaseModule):

    def setup_simulations(self, vunit_proj, **kwargs):
        tb = vunit_proj.library(self.library_name).test_bench("tb_unsigned_divider")
        for dividend_width in [4, 7, 8]:
            for divisor_width in [4, 7, 8]:
                name = "%i_div_%i" % (dividend_width, divisor_width)
                tb.add_config(name=name, generics=dict(dividend_width=dividend_width, divisor_width=divisor_width))



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from pathlib import Path
import os
import unittest

import pytest

from tsfpga.git_utils import (check_that_git_commands_are_available,
                              find_git_files,
                              get_git_commit,
                              git_local_changes_are_present)
from tsfpga.system_utils import create_file, delete, run_command, system_is_windows


THIS_FILE = Path(__file__)
THIS_DIR = THIS_FILE.parent


def test_this_file_is_listed_by_find_git_files_without_argument():
    git_files = find_git_files()
    assert THIS_FILE in git_files


def test_this_file_is_listed_by_find_git_files_with_argument():
    # Test with string as well as tuple
    git_files = find_git_files(file_endings_include="py")
    assert THIS_FILE in git_files
    git_files = find_git_files(file_endings_include=("py", ))
    assert THIS_FILE in git_files


def test_this_file_is_not_listed_by_find_git_files_with_bad_argument():
    git_files = find_git_files(file_endings_include="vhd")
    assert THIS_FILE not in git_files


def test_this_file_is_not_listed_by_find_git_files_with_file_endings_avoid():
    # Test with string as well as tuple
    git_files = find_git_files(file_endings_avoid="py")
    assert THIS_FILE not in git_files
    git_files = find_git_files(file_endings_avoid=("py", ))
    assert THIS_FILE not in git_files


def test_this_file_is_not_listed_by_find_git_files_with_exclude_directory():
    git_files = find_git_files(exclude_directories=[THIS_DIR])
    assert THIS_FILE not in git_files

    git_files = find_git_files(exclude_directories=[THIS_DIR.parent])
    assert THIS_FILE not in git_files

    git_files = find_git_files(exclude_directories=[THIS_DIR.parent, THIS_DIR])
    assert THIS_FILE not in git_files

    git_files = find_git_files(exclude_directories=[THIS_FILE])
    assert THIS_FILE not in git_files


def test_this_file_is_listed_by_find_git_files_with_bad_exclude_directory():
    git_files = find_git_files(exclude_directories=[THIS_DIR.parent / "apa"])
    assert THIS_FILE in git_files


def test_check_that_git_commands_are_available_with_current_cwd_should_pass():
    check_that_git_commands_are_available()


def test_check_that_git_commands_are_available_with_invalid_cwd_should_raise_exception():
    if system_is_windows():
        path_outside_of_repo = "c:/"
    else:
        path_outside_of_repo = "/"

    with pytest.raises(RuntimeError) as exception_info:
        check_that_git_commands_are_available(cwd=path_outside_of_repo)
    assert str(exception_info.value).startswith("Could not run git")


def test_get_git_commit_without_local_changes():
    if git_local_changes_are_present():
        assert False, "Must be run from clean repo (on CI server or locally)"
    assert len(get_git_commit()) == 16


class TestGitCommitWithLocalChanges(unittest.TestCase):

    _trash_file = "local_file_for_git_test.apa"
    _trash_file_path = THIS_DIR / _trash_file

    def setUp(self):
        check_that_git_commands_are_available(cwd=THIS_DIR)
        create_file(self._trash_file_path)
        run_command(["git", "add", self._trash_file], cwd=THIS_DIR)

    def tearDown(self):
        run_command(["git", "reset", self._trash_file], cwd=THIS_DIR)
        delete(self._trash_file_path)

    @staticmethod
    def test_get_git_commit_with_local_changes():
        assert get_git_commit().endswith(" (local changes present)")

    @staticmethod
    def test_get_git_commit_with_env_variable_and_local_changes():
        if "GIT_COMMIT" in os.environ:
            old_git_commit = os.environ["GIT_COMMIT"]
        else:
            old_git_commit = None

        os.environ["GIT_COMMIT"] = "54849b5a8152b07e0809b8f90fc24d54262cb4d6"
        assert get_git_commit() == os.environ["GIT_COMMIT"][0:16]

        if old_git_commit is None:
            del os.environ["GIT_COMMIT"]
        else:
            os.environ["GIT_COMMIT"] = old_git_commit



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

import pytest

import tsfpga
from tsfpga.git_utils import find_git_files
from tsfpga.system_utils import create_file


def open_file_with_encoding(file):
    print(file)
    with file.open(encoding="ascii") as file_handle:
        file_handle.read()


def test_all_checked_in_files_are_properly_encoded():
    """
    To avoid problems with different editors and different file encodings, all checked in files should
    contain only ASCII characters.

    Avoid one of the documentation files that uses wonky characters to illustrate a directory tree.
    """
    for file in files_to_test(exclude_directories=[tsfpga.TSFPGA_DOC / "sphinx" / "module_structure.rst"]):
        open_file_with_encoding(file)


def check_file_ends_with_newline(file):
    test_ok = True
    with file.open() as file_handle:
        file_data = file_handle.read()
        if len(file_data) != 0:
            if file_data[-1] != "\n":
                print(f"File {file} didn't end with newline")
                test_ok = False
    return test_ok


def test_all_checked_in_files_end_with_newline():
    """
    All checked in non-empty files should end with a UNIX style line break (\n).
    Otherwise UNIX doesn't consider them actual text files.
    """
    test_ok = True
    for file in files_to_test():
        test_ok &= check_file_ends_with_newline(file)
    assert test_ok


def check_file_for_tab_character(file):
    test_ok = True
    with file.open() as file_handle:
        for idx, line in enumerate(file_handle.readlines()):
            if "\t" in line:
                test_ok = False
                print("TAB charatcher (\\t) on line %i in %s" % (idx + 1, file))
    return test_ok


def test_no_checked_in_files_contain_tabs():
    """
    To avoid problems with files looking different in different editors, no checked in files may
    contain TAB characters.
    """
    test_ok = True
    for file in files_to_test():
        test_ok &= check_file_for_tab_character(file)
    assert test_ok


def check_file_for_carriage_return(file):
    test_ok = True
    with file.open(newline="") as file_handle:
        if "\r" in file_handle.read():
            test_ok = False
            print(f"Windows style line breaks (\\r\\n aka CR/LF) in {file}")
    return test_ok


def test_no_checked_in_files_contain_carriage_return():
    """
    All checked in files should use UNIX style line breaks (\n not \r\n).
    Some Linux editors and tools will display or interpret the \r as something other than a line break.
    """
    test_ok = True
    for file in files_to_test():
        test_ok &= check_file_for_carriage_return(file)
    assert test_ok


def check_file_for_trailing_whitespace(file):
    test_ok = True
    with file.open() as file_handle:
        for idx, line in enumerate(file_handle.readlines()):
            if " \n" in line:
                test_ok = False
                print("Trailing whitespace on line %i in %s" % (idx + 1, file))
    return test_ok


def test_no_checked_in_files_contain_trailing_whitespace():
    """
    Trailing whitespace is not allowed.
    Some motivation here: https://softwareengineering.stackexchange.com/questions/121555/why-is-trailing-whitespace-a-big-deal
    """
    test_ok = True
    for file in files_to_test():
        test_ok &= check_file_for_trailing_whitespace(file)
    assert test_ok


def files_to_test(exclude_directories=None):
    # Do not test binary image files
    return find_git_files(exclude_directories=exclude_directories, file_endings_avoid="png")


def test_open_file_with_encoding_should_raise_exception_on_bad_file(tmp_path):
    """
    Sanity check that the function we use actually triggers on bad files.
    """
    file = tmp_path / "temp_file_for_test.txt"
    with file.open("w", encoding="utf-8") as file_handle:
        data = "\N{LATIN CAPITAL LETTER O WITH DIAERESIS}"  # Swedish word for island = non-ASCII character
        file_handle.write(data)

    with pytest.raises(UnicodeDecodeError):
        open_file_with_encoding(file)


def test_check_file_for_tab_character_should_fail_on_bad_file(tmp_path):
    """
    Sanity check that the function we use actually triggers on bad files.
    """
    data = "Apa\thest"
    file = create_file(tmp_path / "temp_file_for_test.txt", data)
    assert not check_file_for_tab_character(file)


def test_check_file_for_carriage_return_should_fail_on_bad_file(tmp_path):
    """
    Sanity check that the function we use actually triggers on bad files.
    """
    file = tmp_path / "temp_file_for_test.txt"
    data = b"Apa\r\nhest"
    with file.open("wb") as file_handle:
        file_handle.write(data)
    assert not check_file_for_carriage_return(file)


def test_check_file_for_trailing_whitespace(tmp_path):
    """
    Sanity check that the function we use actually triggers on bad files.
    """
    data = "Apa \nhest    \nzebra"
    file = create_file(tmp_path / "temp_file_for_test.txt", data)
    assert not check_file_for_trailing_whitespace(file)



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

import subprocess
from os import makedirs
from os.path import exists

from tsfpga.sby_writer import SbyWriter


class YosysProject:
    """
    Used for handling a Yosys poject, synthesized with GHDL.
    Currently supports formal verification, but has no real build support.
    """

    def __init__(
            self,
            modules,
            top=None,
            generics=None,
            formal_settings=None,
    ):
        self.modules = modules
        self.top = top
        self.generics = generics
        self.formal_settings = formal_settings

    def run_formal(self, project_path, src_files, compilation_outputs):
        if not exists(project_path):
            makedirs(project_path)

        run_symbiyosys_sby = project_path / "run_symbiyosys.sby"
        SbyWriter.write_sby(
            run_symbiyosys_sby,
            self.top,
            self.generics,
            self.formal_settings,
            compilation_outputs,
            src_files)

        sby_cmd = ["sby", "--yosys", "yosys -m ghdl", "-f", str(run_symbiyosys_sby)]
        return subprocess.call(sby_cmd)



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

import re
import subprocess
from pathlib import Path

from tsfpga.system_utils import create_directory
from .common import get_vivado_path
from .simlib_common import VivadoSimlibCommon


class VivadoSimlibGhdl(VivadoSimlibCommon):

    """
    Handle Vivado simlib with GHDL.
    """

    _libraries = ["unisim", "secureip", "unimacro", "unifast"]

    def __init__(self, output_path, vunit_proj, simulator_interface, vivado_path):
        """
        Args:
            output_path (`pathlib.Path`): The compiled simlib will be placed here.
            vunit_proj: The VUnit project that is used to run simulation.
            simulator_interface: A VUnit SimulatorInterface class.
            vivado_path (`pathlib.Path`): Path to Vivado executable.
        """
        self._vunit_proj = vunit_proj
        self._vivado_path = get_vivado_path(vivado_path)

        self.ghdl_binary = Path(simulator_interface.find_prefix()) / "ghdl"

        self._output_path = output_path / self._get_version_tag()

    def _compile(self):
        vivado_libraries_path = self._vivado_path.parent.parent / "data" / "vhdl" / "src"

        self._compile_unisim(vivado_libraries_path / "unisims")
        self._compile_secureip(vivado_libraries_path / "unisims" / "secureip")
        self._compile_unimacro(vivado_libraries_path / "unimacro")
        self._compile_unifast(vivado_libraries_path / "unifast" / "primitive")

    def _compile_unisim(self, library_path):
        for vhd_file_base in ["unisim_VPKG", "unisim_VCOMP", "retarget_VCOMP", "unisim_retarget_VCOMP"]:
            vhd_file = library_path / (vhd_file_base + ".vhd")
            assert vhd_file.exists, vhd_file
            self._compile_ghdl_file(vhd_file, "unisim")

        primitive_dir = library_path / "primitive"
        for vhd_file in self._iterate_compile_order(primitive_dir):
            self._compile_ghdl_file(vhd_file, "unisim")

        retarget_dir = library_path / "retarget"
        for vhd_file in retarget_dir.glob("*.vhd"):
            self._compile_ghdl_file(vhd_file, "unisim")

    def _compile_secureip(self, library_path):
        for vhd_file in library_path.glob("*.vhd"):
            self._compile_ghdl_file(vhd_file, "secureip")

    def _compile_unimacro(self, library_path):
        vhd_file = library_path / "unimacro_VCOMP.vhd"
        assert vhd_file.exists, vhd_file
        self._compile_ghdl_file(vhd_file, "unimacro")

        for vhd_file in self._iterate_compile_order(library_path):
            self._compile_ghdl_file(vhd_file, "unimacro")

    def _compile_unifast(self, library_path):
        for vhd_file in self._iterate_compile_order(library_path):
            self._compile_ghdl_file(vhd_file, "unifast")

    @staticmethod
    def _iterate_compile_order(library_path):
        with (library_path / "vhdl_analyze_order").open() as file_handle:
            for vhd_file_base in file_handle.readlines():
                vhd_file = library_path / vhd_file_base.strip()
                assert vhd_file.exists(), vhd_file
                yield vhd_file

    def _compile_ghdl_file(self, vhd_file, library_name):
        print(f"Compiling {vhd_file} into {library_name}")
        workdir = self._output_path / library_name / "v08"
        create_directory(workdir, empty=False)
        cmd = [
            self.ghdl_binary,
            "-a",
            "--ieee=synopsys",
            "--std=08",
            "--workdir=" + str(workdir.resolve()),
            "-P" + str(self._output_path),
            "-fexplicit",
            "-frelaxed-rules",
            "--no-vital-checks",
            "--warn-binding",
            "--mb-comments",
            "--work=" + library_name,
            str(vhd_file.resolve())
        ]
        subprocess.check_call(cmd, cwd=self._output_path)

    def _get_simulator_tag(self):
        """
        Return simulator version tag
        """
        cmd = [self.ghdl_binary, "--version"]
        output = subprocess.check_output(cmd).decode()

        regexp_with_tag = re.compile(r"^GHDL (\S+) \((\S+)\).*")
        match = regexp_with_tag.search(output)
        if match is not None:
            return self._format_version(f"ghdl_{match.group(1)}_{match.group(2)}")

        regexp_without_tag = re.compile(r"^GHDL (\S+).*")
        match = regexp_without_tag.search(output)
        if match is not None:
            return self._format_version("ghdl_" + match.group(1))

        raise ValueError("Could not find GHDL version string: " + output)

    def _add_to_vunit_project(self):
        """
        Add the compiled simlib to your VUnit project.
        """
        for library in self._libraries:
            library_path = self._output_path / library / "v08"
            assert library_path.exists(), library_path
            self._vunit_proj.add_external_library(library, library_path)



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from collections import OrderedDict
from pathlib import Path
import unittest

import pytest

from tsfpga.build_step_tcl_hook import BuildStepTclHook
from tsfpga.module import get_modules
from tsfpga.system_utils import create_file
from tsfpga.vivado.common import to_tcl_path
from tsfpga.vivado.tcl import VivadoTcl
from tsfpga.test import file_contains_string
from tsfpga.test.conftest import fixture_tmp_path  # pylint: disable=unused-import


def test_set_create_run_index():
    tcl = VivadoTcl(name="").create(
        project_folder=Path(),
        modules=[],
        part="",
        top="",
        run_index=2
    )
    assert "\ncurrent_run [get_runs synth_2]\n" in tcl


def test_static_generics():
    # Use OrderedDict here in test so that order will be preserved and we can test for equality.
    # In real world case a normal dict can be used.
    generics = OrderedDict(enable=True, disable=False, integer=123, slv="4'b0101")

    tcl = VivadoTcl(name="").create(
        project_folder=Path(),
        modules=[],
        part="",
        top="",
        run_index=1,
        generics=generics
    )
    expected = "\nset_property generic {enable=1'b1 disable=1'b0 integer=123 slv=4'b0101} [current_fileset]\n"
    assert expected in tcl


def test_build_step_hooks():
    dummy = BuildStepTclHook(Path("dummy.tcl"), "STEPS.SYNTH_DESIGN.TCL.PRE")
    files = BuildStepTclHook(Path("files.tcl"), "STEPS.ROUTE_DESIGN.TCL.PRE")
    tcl = VivadoTcl(name="").create(
        project_folder=Path(),
        modules=[],
        part="part",
        top="",
        run_index=1,
        build_step_hooks=[dummy, files]
    )

    assert f"\nset_property STEPS.SYNTH_DESIGN.TCL.PRE {to_tcl_path(dummy.tcl_file)} ${{run}}\n" in tcl
    assert f"\nset_property STEPS.ROUTE_DESIGN.TCL.PRE {to_tcl_path(files.tcl_file)} ${{run}}\n" in tcl


def test_build_step_hooks_with_same_hook_step(tmp_path):
    dummy = BuildStepTclHook(Path("dummy.tcl"), "STEPS.SYNTH_DESIGN.TCL.PRE")
    files = BuildStepTclHook(Path("files.tcl"), "STEPS.SYNTH_DESIGN.TCL.PRE")
    tcl = VivadoTcl(name="").create(
        project_folder=tmp_path / "dummy_project_folder",
        modules=[],
        part="part",
        top="",
        run_index=1,
        build_step_hooks=[dummy, files]
    )

    hook_file = tmp_path / "dummy_project_folder" / "hook_STEPS_SYNTH_DESIGN_TCL_PRE.tcl"

    assert file_contains_string(hook_file, f"source {to_tcl_path(dummy.tcl_file)}")
    assert file_contains_string(hook_file, f"source {to_tcl_path(files.tcl_file)}")

    assert f"\nset_property STEPS.SYNTH_DESIGN.TCL.PRE {to_tcl_path(hook_file)} ${{run}}\n" in tcl


def test_ip_cache_location(tmp_path):
    tcl = VivadoTcl(name="").create(
        project_folder=Path(),
        modules=[],
        part="part",
        top="",
        run_index=1
    )
    assert "config_ip_cache" not in tcl

    tcl = VivadoTcl(name="").create(
        project_folder=Path(),
        modules=[],
        part="part",
        top="",
        run_index=1,
        ip_cache_path=tmp_path
    )
    assert f"\nconfig_ip_cache -use_cache_location {to_tcl_path(tmp_path)}\n" in tcl


def test_set_multiple_threads():
    num_threads = 2
    tcl = VivadoTcl(name="").build(
        project_file=Path(),
        output_path=Path(),
        num_threads=num_threads,
        run_index=1
    )
    assert "set_param general.maxThreads %d" % num_threads in tcl
    assert "launch_runs synth_1 -jobs %d" % num_threads in tcl
    assert "launch_runs impl_1 -jobs %d" % num_threads in tcl


def test_set_build_run_index():
    tcl = VivadoTcl(name="").build(
        project_file=Path(),
        output_path=Path(),
        num_threads=0,
        run_index=1
    )
    assert "impl_1" in tcl
    assert "synth_1" in tcl
    assert "impl_2" not in tcl
    assert "synth_2" not in tcl

    tcl = VivadoTcl(name="").build(
        project_file=Path(),
        output_path=Path(),
        num_threads=0,
        run_index=2
    )
    assert "impl_2" in tcl
    assert "synth_2" in tcl
    assert "impl_1" not in tcl
    assert "synth_1" not in tcl


def test_runtime_generics():
    tcl = VivadoTcl(name="").build(
        project_file=Path(),
        output_path=Path(),
        num_threads=0,
        run_index=0,
        generics=dict(dummy=True)
    )
    expected = "\nset_property generic {dummy=1'b1} [current_fileset]\n"
    assert expected in tcl


# pylint: disable=too-many-instance-attributes
@pytest.mark.usefixtures("fixture_tmp_path")
class TestVivadoTcl(unittest.TestCase):

    tmp_path = None

    def setUp(self):
        self.modules_folder = self.tmp_path / "modules"

        # A library with some synth files and some test files
        self.a_vhd = to_tcl_path(create_file(self.modules_folder / "apa" / "a.vhd"))
        self.tb_a_vhd = to_tcl_path(create_file(self.modules_folder / "apa" / "test" / "tb_a.vhd"))
        self.a_xdc = to_tcl_path(create_file(self.modules_folder / "apa" / "scoped_constraints" / "a.xdc"))

        self.b_v = to_tcl_path(create_file(self.modules_folder / "apa" / "b.v"))
        self.b_tcl = to_tcl_path(create_file(self.modules_folder / "apa" / "scoped_constraints" / "b.tcl"))

        self.c_tcl = to_tcl_path(create_file(self.modules_folder / "apa" / "ip_cores" / "c.tcl"))

        # A library with only test files
        self.c_vhd = to_tcl_path(create_file(self.modules_folder / "zebra" / "test" / "c.vhd"))

        self.modules = get_modules([self.modules_folder])

        self.tcl = VivadoTcl(name="name")

    def test_only_synthesis_files_added_to_create_project_tcl(self):
        tcl = self.tcl.create(
            project_folder=Path(),
            modules=self.modules,
            part="",
            top="",
            run_index=1
        )
        assert self.a_vhd in tcl and self.b_v in tcl
        assert self.tb_a_vhd not in tcl and "tb_a.vhd" not in tcl

    def test_different_hdl_file_types(self):
        tcl = self.tcl.create(
            project_folder=Path(),
            modules=self.modules,
            part="",
            top="",
            run_index=1
        )
        assert f"read_vhdl -library apa -vhdl2008 {{{self.a_vhd}}}" in tcl
        assert f"read_verilog {{{self.b_v}}}" in tcl

    def test_constraints(self):
        tcl = self.tcl.create(
            project_folder=Path(),
            modules=self.modules,
            part="part",
            top="",
            run_index=1
        )

        expected = "\nread_xdc -ref a %s\n" % self.a_xdc
        assert expected in tcl
        expected = "\nread_xdc -ref b -unmanaged %s\n" % self.b_tcl
        assert expected in tcl

    def test_ip_core_files(self):
        tcl = self.tcl.create(
            project_folder=Path(),
            modules=self.modules,
            part="part",
            top="",
            run_index=1
        )
        assert "\nsource -notrace %s\n" % self.c_tcl in tcl

    def test_empty_library_not_in_create_project_tcl(self):
        tcl = self.tcl.create(
            project_folder=Path(),
            modules=self.modules,
            part="part",
            top="",
            run_index=1
        )
        assert "zebra" not in tcl

    def test_multiple_tcl_sources(self):
        extra_tcl_sources = [Path("dummy.tcl"), Path("files.tcl")]
        tcl = self.tcl.create(
            project_folder=Path(),
            modules=self.modules,
            part="part",
            top="",
            run_index=1,
            tcl_sources=extra_tcl_sources
        )

        for filename in extra_tcl_sources:
            assert f"\nsource -notrace {to_tcl_path(filename)}\n" in tcl

    def test_io_buffer_setting(self):
        tcl = self.tcl.create(
            project_folder=Path(),
            modules=self.modules,
            part="part",
            top="",
            run_index=1,
            disable_io_buffers=True
        )

        no_io_buffers_tcl = "\nset_property -name {STEPS.SYNTH_DESIGN.ARGS.MORE OPTIONS} -value -no_iobuf -objects [get_runs synth_1]\n"
        assert no_io_buffers_tcl in tcl

        tcl = self.tcl.create(
            project_folder=Path(),
            modules=self.modules,
            part="part",
            top="",
            run_index=1,
            disable_io_buffers=False
        )

        assert no_io_buffers_tcl not in tcl



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

import pytest


@pytest.fixture
def fixture_tmp_path(request, tmp_path):
    """
    A pytest fixture for usage in unittest.TestCase style test classes which gives
    access to a unique temp path for each test case.
    """
    # Set class member tmp_path
    request.cls.tmp_path = tmp_path



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

"""
A set of reusable functions for working with HDL projects.
"""


from pathlib import Path


THIS_DIR = Path(__file__).parent
REPO_ROOT = (THIS_DIR / "..").resolve()

TSFPGA_PATH = REPO_ROOT / "tsfpga"
TSFPGA_DOC = REPO_ROOT / "doc"
TSFPGA_MODULES = REPO_ROOT / "modules"
TSFPGA_TCL = THIS_DIR / "vivado" / "tcl"
TSFPGA_GENERATED = REPO_ROOT / "generated"

TSFPGA_EXAMPLES = REPO_ROOT / "examples"
TSFPGA_EXAMPLE_MODULES = TSFPGA_EXAMPLES / "modules"
TSFPGA_EXAMPLE_MODULES_WITH_IP = TSFPGA_EXAMPLES / "modules_with_ip"

ALL_TSFPGA_MODULES_FOLDERS = [TSFPGA_MODULES, TSFPGA_EXAMPLE_MODULES, TSFPGA_EXAMPLE_MODULES_WITH_IP]

__version__ = "4.0.1"



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------


from tsfpga.register_code_generator import RegisterCodeGenerator
from tsfpga.register_types import Register, RegisterArray


class RegisterVhdlGenerator(RegisterCodeGenerator):

    def __init__(self, module_name, generated_info):
        self.module_name = module_name
        self.generated_info = generated_info

    @staticmethod
    def _comment(comment):
        return f"-- {comment}\n"

    def _header(self):
        return self._comment(self.generated_info)

    def _register_name(self, register, register_array=None):
        if register_array is None:
            return f"{self.module_name}_{register.name}"
        return f"{self.module_name}_{register_array.name}_{register.name}"

    def _register_function_signature(self, register, register_array):
        return f"function {self._register_name(register, register_array)}(array_index : natural) return integer"

    def _register_indexes(self, register_objects):
        vhdl = ""
        for register, register_array in self._iterate_registers(register_objects):
            if register_array is None:
                vhdl += f"  constant {self._register_name(register)} : integer := {register.index};\n"
            else:
                vhdl += f"  {self._register_function_signature(register, register_array)};\n"

        return vhdl

    def _array_length_constant_name(self, register_array):
        return f"{self.module_name}_{register_array.name}_array_length"

    def _array_constants(self, register_objects):
        vhdl = ""
        for register_object in register_objects:
            if isinstance(register_object, RegisterArray):
                constant = self._array_length_constant_name(register_object)
                vhdl += f"  constant {constant} : integer := {register_object.length};\n"
        if vhdl:
            vhdl += "\n"

        return vhdl

    def _register_map(self, register_objects):
        map_name = f"{self.module_name}_reg_map"

        register_definitions = []
        default_values = []
        for register_object in register_objects:
            if isinstance(register_object, Register):
                idx = self._register_name(register_object)
                register_definitions.append(f"  (idx => {idx}, reg_type => {register_object.mode})")
                default_values.append(f"  std_logic_vector(to_signed({register_object.default_value}, 32))")
            else:
                for array_index in range(register_object.length):
                    for register in register_object.registers:
                        idx = f"{self._register_name(register, register_object)}({array_index})"
                        register_definitions.append(f"  (idx => {idx}, reg_type => {register.mode})")
                        default_values.append(f"  std_logic_vector(to_signed({register.default_value}, 32))")

        last_index = register_objects[-1].index
        vhdl = f"  constant {map_name} : reg_definition_vec_t(0 to {last_index}) := (\n  "
        vhdl += ",\n  ".join(register_definitions)
        vhdl += "\n  );\n"
        vhdl += "\n"
        vhdl += f"  subtype {self.module_name}_regs_t is reg_vec_t({map_name}'range);\n"
        vhdl += f"  subtype {self.module_name}_reg_was_written_t is std_logic_vector({map_name}'range);\n"
        vhdl += f"  constant {self.module_name}_regs_init : {self.module_name}_regs_t := (\n  "
        vhdl += ",\n  ".join(default_values)
        vhdl += "\n  );\n"
        return vhdl

    def _register_bits(self, register_objects):
        vhdl = ""
        for register, register_array in self._iterate_registers(register_objects):
            for bit in register.bits:
                name = f"{self._register_name(register, register_array)}_{bit.name}"
                vhdl += f"  constant {name} : integer := {bit.index};\n"
            if register.bits:
                vhdl += "\n"

        return vhdl

    def _array_index_functions(self, register_objects):
        vhdl = ""
        for register_object in register_objects:
            if isinstance(register_object, RegisterArray):
                num_registers = len(register_object.registers)
                array_length = self._array_length_constant_name(register_object)
                for register in register_object.registers:
                    vhdl += f"  {self._register_function_signature(register, register_object)} is\n"
                    vhdl += "  begin\n"
                    vhdl += f"    assert array_index < {array_length} \n"
                    vhdl += "      report \"Array index out of bounds: \" & to_string(array_index)\n"
                    vhdl += "      severity failure;\n"
                    vhdl += f"    return {register_object.base_index} + array_index * {num_registers} + {register.index};\n"
                    vhdl += "  end function;\n\n"

        return vhdl

    def _constants(self, constants):
        vhdl = ""
        for constant in constants:
            vhdl += f"  constant {self.module_name}_constant_{constant.name} : integer := {constant.value};\n"
        return vhdl

    def get_package(self, register_objects, constants):
        pkg_name = f"{self.module_name}_regs_pkg"

        vhdl = f"""\
{self._header()}
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library reg_file;
use reg_file.reg_file_pkg.all;


package {pkg_name} is

{self._register_indexes(register_objects)}
{self._array_constants(register_objects)}\
{self._register_map(register_objects)}
{self._register_bits(register_objects)}\
{self._constants(constants)}
end package;

package body {pkg_name} is

{self._array_index_functions(register_objects)}\
end package body;
"""

        return vhdl



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

import datetime
from shutil import copy2

import toml

from tsfpga.git_utils import git_commands_are_available, get_git_commit
from tsfpga.register_types import Constant, Register, RegisterArray
from tsfpga.register_c_generator import RegisterCGenerator
from tsfpga.register_cpp_generator import RegisterCppGenerator
from tsfpga.register_html_generator import RegisterHtmlGenerator
from tsfpga.register_vhdl_generator import RegisterVhdlGenerator
from tsfpga.svn_utils import svn_commands_are_available, get_svn_revision_information
from tsfpga.system_utils import create_directory, create_file, read_file


class RegisterList:

    """
    Used to handle the registers of a module. Also known as a register map.
    """

    def __init__(self, name, source_definition_file):
        """
        Args:
            name (str): The name of this register list. Typically the name of the module that uses it.
            source_definition_file (`pathlib.Path`): The TOML source file that defined this register list.
        """
        self.name = name
        self.source_definition_file = source_definition_file
        self.register_objects = []
        self.constants = []

    def append_register(self, name, mode, description=None, default_value=None):
        """
        Append a register to this list.

        Args:
            name (str): The name of the register.
            mode (str): A valid register mode.
            description (str): Textual register description.
            default_value (int): Default value for the register (signed).
        Return:
            :class:`.Register`: The register object that was created.
        """
        if self.register_objects:
            index = self.register_objects[-1].index + 1
        else:
            index = 0

        register = Register(name, index, mode, description, default_value)
        self.register_objects.append(register)

        return register

    def append_register_array(self, name, length):
        """
        Append a register array to this list.

        Args:
            name (str): The name of the register array.
            length (int): The number of times the register sequence shall be repeated.
        Return:
            :class:`.RegisterArray`: The register array object that was created.
        """
        if self.register_objects:
            base_index = self.register_objects[-1].index + 1
        else:
            base_index = 0
        register_array = RegisterArray(name, base_index, length)

        self.register_objects.append(register_array)
        return register_array

    def get_register(self, name):
        """
        Get a register from this list. Will only find single registers, not registers in a
        register array.

        Args:
            name (str): The name of the register.
        Return:
            :class:`.Register`: The register. ``None`` if no register matched.
        """
        for register_object in self.register_objects:
            if register_object.name == name:
                return register_object

        return None

    def add_constant(self, name, value):
        """
        Add a constant. Will be available in the generated packages and headers.

        Args:
            name (str): The name of the constant.
            length (int): The constant value (signed).
        """
        self.constants.append(Constant(name, value))

    def create_vhdl_package(self, output_path):
        """
        Create a VHDL package file with register and field definitions.

        This function assumes that the ``output_path`` folder already exists.
        This assumption makes it slightly faster than the other functions that use ``create_file()``.
        Necessary since this one is often used in real time (before simulations, etc..) and not in
        one-off scenarios like the others (when making a release).

        Args:
            output_path (`pathlib.Path`): Result will be placed here.
        """
        register_vhdl_generator = RegisterVhdlGenerator(self.name, self.generated_info())
        with (output_path / (self.name + "_regs_pkg.vhd")).open("w") as file_handle:
            file_handle.write(register_vhdl_generator.get_package(self.register_objects, self.constants))

    def create_c_header(self, output_path):
        """
        Create a C header file with register and field definitions.

        Args:
            output_path (`pathlib.Path`): Result will be placed here.
        """
        output_file = output_path / (self.name + "_regs.h")
        register_c_generator = RegisterCGenerator(self.name, self.generated_source_info())
        create_file(output_file, register_c_generator.get_header(self.register_objects, self.constants))

    def create_cpp_interface(self, output_path):
        """
        Create a C++ class interface header file, with register and field definitions. The interface header
        contains only virtual methods.

        Args:
            output_path (`pathlib.Path`): Result will be placed here.
        """
        output_file = output_path / ("i_" + self.name + ".h")
        register_cpp_generator = RegisterCppGenerator(self.name, self.generated_source_info())
        create_file(output_file, register_cpp_generator.get_interface(self.register_objects, self.constants))

    def create_cpp_header(self, output_path):
        """
        Create a C++ class header file.

        Args:
            output_path (`pathlib.Path`): Result will be placed here.
        """
        output_file = output_path / (self.name + ".h")
        register_cpp_generator = RegisterCppGenerator(self.name, self.generated_source_info())
        create_file(output_file, register_cpp_generator.get_header(self.register_objects))

    def create_cpp_implementation(self, output_path):
        """
        Create a C++ class implementation file.

        Args:
            output_path (`pathlib.Path`): Result will be placed here.
        """
        output_file = output_path / (self.name + ".cpp")
        register_cpp_generator = RegisterCppGenerator(self.name, self.generated_source_info())
        create_file(output_file, register_cpp_generator.get_implementation(self.register_objects))

    def create_html_page(self, output_path):
        """
        Create a documentation HTML page with register and field information. Will include the
        table created by :meth:`.create_html_table`.

        Args:
            output_path (`pathlib.Path`): Result will be placed here.
        """
        output_file = output_path / (self.name + "_regs.html")
        register_html_generator = RegisterHtmlGenerator(self.name, self.generated_source_info())
        create_file(output_file, register_html_generator.get_page(self.register_objects))

    def create_html_table(self, output_path):
        """
        Create documentation HTML table with register and field information.

        Args:
            output_path (`pathlib.Path`): Result will be placed here.
        """
        output_file = output_path / (self.name + "_regs_table.html")
        register_html_generator = RegisterHtmlGenerator(self.name, self.generated_source_info())
        create_file(output_file, register_html_generator.get_table(self.register_objects))

    def copy_source_definition(self, output_path):
        """
        Copy the TOML file that created this register list.

        Args:
            output_path (`pathlib.Path`): Result will be placed here.
        """
        create_directory(output_path, empty=False)
        copy2(self.source_definition_file, output_path)

    @staticmethod
    def generated_info():
        """
        Return:
            str: A string informing the user that a file is automatically generated.
        """
        return "File automatically generated by tsfpga."

    def generated_source_info(self):
        """
        Return:
            str: A string informing the user that a file is automatically generated, containing
            info about the source of the generated register information.
        """
        time_info = datetime.datetime.now().strftime("%Y-%m-%d %H:%M")

        commit_info = ""
        if git_commands_are_available():
            commit_info = f" at commit {get_git_commit()}"
        elif svn_commands_are_available():
            commit_info = f" at revision {get_svn_revision_information()}"

        info = f"Generated {time_info} from file {self.source_definition_file.name}{commit_info}."

        return f"{self.generated_info()} {info}"


def get_default_registers():
    """
    tsfpga default registers
    """
    registers = [
        Register("config", 0, "r_w", "Configuration register."),
        Register(
            "command", 1, "wpulse",
            "When this register is written, all '1's in the written word will be asserted for one "
            "clock cycle in the FPGA logic."),
        Register("status", 2, "r", "Status register."),
        Register(
            "irq_status", 3, "r_wpulse",
            "Reading a '1' in this register means the corresponding interrupt has triggered. Writing "
            "to this register will clear the interrupts where there is a '1' in the written word."),
        Register(
            "irq_mask", 4, "r_w",
            "A '1' in this register means that the corresponding interrupt is enabled. ")
    ]
    return registers


def load_toml_file(toml_file):
    if not toml_file.exists():
        raise FileNotFoundError(f"Requested TOML file does not exist: {toml_file}")

    raw_toml = read_file(toml_file)
    try:
        return toml.loads(raw_toml)
    except toml.TomlDecodeError as exception_info:
        message = f"Error while parsing TOML file {toml_file}:\n{exception_info}"
        raise ValueError(message) from exception_info


def from_toml(module_name, toml_file, default_registers=None):
    toml_data = load_toml_file(toml_file)
    register_list = RegisterList(module_name, toml_file)

    default_register_names = []
    if default_registers is not None:
        register_list.register_objects = default_registers
        for register in default_registers:
            default_register_names.append(register.name)

    names_taken = set()

    if "register" in toml_data:
        for name, items in toml_data["register"].items():
            _parse_plain_register(
                name, items, register_list, default_register_names, names_taken, toml_file)

    if "register_array" in toml_data:
        for name, items in toml_data["register_array"].items():
            _parse_register_array(name, items, register_list, names_taken, toml_file)

    return register_list


RECOGNIZED_REGISTER_ITEMS = {"mode", "default_value", "description", "bits"}
RECOGNIZED_REGISTER_ARRAY_ITEMS = {"array_length", "register"}

# pylint: disable=too-many-arguments


def _parse_plain_register(name, items, register_list, default_register_names, names_taken, toml_file):
    if "array_length" in items:
        message = f"Plain register {name} in {toml_file} can not have array_length attribute"
        raise ValueError(message)

    if name in default_register_names:
        # Default registers can be "updated" in the sense that the user can use a custom
        # description and add whatever bits they use in the current module. They can not however
        # change the mode.
        register = register_list.get_register(name)
        if "mode" in items:
            message = f"Overloading register {name} in {toml_file}, one can not change mode from default"
            raise ValueError(message)
    else:
        # If it is a new register however the mode has to be specified.
        if "mode" not in items:
            raise ValueError(f"Register {name} in {toml_file} does not have mode field")
        register = register_list.append_register(name, items["mode"])

    for item_name, item_value in items.items():
        if item_name not in RECOGNIZED_REGISTER_ITEMS:
            message = f"Error while parsing register {name} in {toml_file}:\nUnknown key {item_name}"
            raise ValueError(message)
        if item_name == "default_value":
            register.default_value = item_value

        if item_name == "description":
            register.description = item_value

        if item_name == "bits":
            for bit_name, bit_description in item_value.items():
                register.append_bit(bit_name, bit_description)

    names_taken.add(name)


# pylint: disable=too-many-locals
def _parse_register_array(name, items, register_list, names_taken, toml_file):
    if name in names_taken:
        message = f"Duplicate name {name} in {toml_file}"
        raise ValueError(message)
    if "array_length" not in items:
        message = f"Register array {name} in {toml_file} does not have array_length attribute"
        raise ValueError(message)

    for item_name in items:
        if item_name not in RECOGNIZED_REGISTER_ARRAY_ITEMS:
            message = f"Error while parsing register array {name} in {toml_file}:\n"\
                f"Unknown key {item_name}"
            raise ValueError(message)

    length = items["array_length"]
    register_array = register_list.append_register_array(name, length)

    for register_name, register_items in items["register"].items():
        if "mode" not in register_items:
            message = f"Register {register_name} within array {name} in {toml_file} does not have mode field"
            raise ValueError(message)
        register = register_array.append_register(register_name, register_items["mode"])

        for register_item_name, register_item_value in register_items.items():
            if register_item_name not in RECOGNIZED_REGISTER_ITEMS:
                message = f"Error while parsing register {register_name} in array {name} in " \
                    f"{toml_file}:\nUnknown key {register_item_name}"
                raise ValueError(message)
            if register_item_name == "default_value":
                register.default_value = register_item_value

            if register_item_name == "description":
                register.description = register_item_value

            if register_item_name == "bits":
                for bit_name, bit_description in register_item_value.items():
                    register.append_bit(bit_name, bit_description)



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------


class Mode:

    def __init__(self, mode_readable, description):
        self.mode_readable = mode_readable
        self.description = description


REGISTER_MODES = dict(
    r=Mode("Read", "Bus can read a value that fabric provides."),
    w=Mode("Write", "Bus can write a value that is available for fabric usage."),
    r_w=Mode("Read, Write",
             "Bus can write a value and read it back. The written value is available for fabric usage."),
    wpulse=Mode("Write-pulse",
                "Bus can write a value that is asserted for one clock cycle in fabric."),
    r_wpulse=Mode("Read, Write-pulse",
                  "Bus can read a value that fabric provides. "
                  "Bus can write a value that is asserted for one clock cycle in fabric."),
)


class Bit:

    """
    Used to represent a bit in a register.
    """

    def __init__(self, name, index, description):
        """
        Args:
            name (str): The name of the bit.
            index (int): The zero-based index of this bit within the register.
            description (str): Textual bit description.
        """
        self.name = name
        self.index = index
        self.description = description


class Register:

    """
    Used to represent a register and its fields.
    """

    # pylint: disable=too-many-arguments
    def __init__(self, name, index, mode, description=None, default_value=None):
        """
        Args:
            name (str): The name of the register.
            index (int): The zero-based index of this register in its register list.
            mode (str): A valid register mode.
            description (str): Textual register description.
            default_value (int): Default value for the register (signed).
        """
        if mode not in REGISTER_MODES:
            raise ValueError(f'Invalid mode "{mode}" for register "{name}"')

        self.name = name
        self.index = index
        self.mode = mode
        self.description = "" if description is None else description
        self.default_value = 0 if default_value is None else default_value
        self.bits = []

    def append_bit(self, name, description):
        """
        Append a bit to this register.

        Args:
            name (str): The name of the bit.
            description (str): Description of the bit.
        Return:
            :class:`.Bit`: The bit object that was created.
        """
        index = len(self.bits)
        bit = Bit(name, index, description)

        self.bits.append(bit)
        return bit

    @property
    def address(self):
        """
        int: Byte address, within the register list, of this register.
        """
        return 4 * self.index

    @property
    def is_bus_readable(self):
        """
        True if the register is readable by bus. Based on the register type.
        """
        return self.mode in ["r", "r_w", "r_wpulse"]

    @property
    def is_bus_writeable(self):
        """
        True if the register is writeable by bus. Based on the register type.
        """
        return self.mode in ["w", "r_w", "wpulse", "r_wpulse"]


class RegisterArray:

    """
    Represent an array of registers. That is, a sequence of registers that shall be repeated a
    number of times in a register list.
    """

    def __init__(self, name, base_index, length):
        """
        Args:
            name (str): The name of this register array.
            base_index (int): The zero-based index of the first register of this array in
                the register list.
            length (int): The number of times the register sequence shall be repeated.
        """
        self.name = name
        self.base_index = base_index
        self.length = length
        self.registers = []

    def append_register(self, name, mode, description=None, default_value=None):
        """
        Append a register to this array.

        Args:
            name (str): The name of the register.
            mode (str): A valid register mode.
            description (str): Textual register description.
            default_value (int): Default value for the register (signed).
        Return:
            :class:`.Register`: The register object that was created.
        """
        index = len(self.registers)
        register = Register(name, index, mode, description, default_value)

        self.registers.append(register)
        return register

    @property
    def index(self):
        """
        Property exists to be used analogously with ``Register.index``.

        Return:
            int: The highest index occupied by this array.
        """
        return self.base_index + self.length * len(self.registers) - 1


class Constant:

    def __init__(self, name, value):
        self.name = name
        self.value = value



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

import os
from os.path import commonpath
from pathlib import Path
import subprocess

from tsfpga import REPO_ROOT


def get_git_commit(cwd=None):
    """
    Generally, eight to ten characters are more than enough to be unique within a project.
    The linux kernel, one of the largest projects, needs 11.
    https://git-scm.com/book/en/v2/Git-Tools-Revision-Selection#Short-SHA-1
    """
    sha_length = 16
    if "GIT_COMMIT" in os.environ:
        return os.environ["GIT_COMMIT"][0:sha_length]

    check_that_git_commands_are_available(cwd)

    git_commit = get_git_sha(cwd)[0:sha_length]
    if git_local_changes_are_present(cwd):
        git_commit += " (local changes present)"

    return git_commit


def get_git_sha(cwd=None):
    command = ["git", "rev-parse", "HEAD"]
    output = subprocess.check_output(command, cwd=cwd, universal_newlines=True)
    return output


def git_local_changes_are_present(cwd=None):
    command = ["git", "diff-index", "--quiet", "HEAD", "--"]
    try:
        subprocess.check_call(command, cwd=cwd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError:
        return True
    return False


def git_commands_are_available(cwd=None):
    command = ["git", "rev-parse", "--is-inside-work-tree"]
    try:
        subprocess.check_call(command, cwd=cwd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError:
        return False
    return True


def check_that_git_commands_are_available(cwd=None):
    if not git_commands_are_available(cwd):
        mesg = "Could not run git. Is the command available on PATH? Is the script called from a repo?"
        raise RuntimeError(mesg)


def find_git_files(file_endings_include=None,
                   file_endings_avoid=None,
                   directory=REPO_ROOT,
                   exclude_directories=None):
    """
    Args:
        file_endings_include (str or tuple(str)). Only files with these endings will be included.
        file_endings_avoid (str or tuple(str)): String or tuple of strings. Files with these endings will not be included.
        directory (`pathlib.Path`): Search in this directory.
        exclude_directories (list(`pathlib.Path`)): Files in these directories will not be included.
    """
    exclude_directories = [] if exclude_directories is None else \
        [exclude_directory.resolve() for exclude_directory in exclude_directories]

    command = ["git", "ls-files"]
    output = subprocess.check_output(command, cwd=directory, universal_newlines=True)
    ls_files = output.split("\n")

    # subprocess.check_output() returns a trailing "\n". The split() call will make that an empty object at the end of the list.
    ls_files = ls_files[:-1]

    for file in ls_files:
        # git ls-files returns paths relative to the working directory where it's called. Hence we prepend the cwd used.
        # Normpath is necessary in windows where you can get a mix of slashes and backslashes which makes
        # path comparisons sketchy
        file = directory.joinpath(Path(file))
        assert file.exists()  # Make sure concatenation of relative path worked

        if file.is_file():  # "git ls-files" also lists submodule folders
            if (file_endings_include is None or file.name.endswith(file_endings_include)) \
                    and (file_endings_avoid is None or not file.name.endswith(file_endings_avoid)):
                if not _file_is_in_directory(file, exclude_directories):
                    yield file


def list_current_tags(cwd=REPO_ROOT):
    command = ["git", "tag", "--list", "--points-at", "HEAD"]
    return subprocess.check_output(command, cwd=cwd).decode().splitlines()


def list_tags(cwd=REPO_ROOT):
    command = ["git", "tag", "--list"]
    return subprocess.check_output(command, cwd=cwd).decode().splitlines()


def _file_is_in_directory(filename, directories):
    for directory in directories:
        if commonpath([str(filename), str(directory)]) == str(directory):
            return True
    return False



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

import hashlib

from tsfpga.system_utils import create_file, delete, read_file
from .project import VivadoProject


class VivadoIpCores:

    """
    Handle a list of IP core sources. Has a mechanism to detect whether a regenerate of IP files is needed.
    """

    _project_name = "vivado_ip_project"

    def __init__(self, modules, output_path, part_name):
        """
        Args:
            modules (list(:class:`Module <.BaseModule>`)): IP cores from  these modules will be included.
            output_path (`pathlib.Path`): The Vivado project will be placed here.
            part_name (str): Vivado part name to be used for the project.
        """
        self._project_folder = output_path.resolve() / self._project_name
        self._part_name = part_name
        self._hash_file = self._project_folder / "ip_files_hash.txt"

        self._setup(modules)

    @property
    def compile_order_file(self):
        """
        `pathlib.Path`: Path to the generated compile order file.
        """
        return self._project_folder / "compile_order.txt"

    @property
    def vivado_project_sources_directory(self):
        """
        `pathlib.Path`: Path to the "sources" directory of the Vivado project.
        """
        return self._project_folder / "vivado_ip_project.srcs" / "sources_1"

    @property
    def vivado_project_file(self):
        """
        `pathlib.Path`: Path to the Vivado project file.
        """
        return self._vivado_project.project_file(self._project_folder)

    def create_vivado_project(self):
        """
        Create IP core Vivado project.
        """
        print(f"Creating IP core project in {self._project_folder}")
        delete(self._project_folder)
        self._vivado_project.create(self._project_folder)
        self._save_hash()

    def create_vivado_project_if_needed(self):
        """
        Create IP core Vivado project if anything has changed since last time this was run.
        If

        * List of TCL files that create IP cores,
        * and contents of these files,

        is the same then it will not create. But if anything is added or removed from the list,
        or the contents of a TCL file is changed, there will be a recreation.

        Return:
            True if Vivado project was created. False otherwise.
        """
        if self._should_create():
            self.create_vivado_project()
            return True

        return False

    def _setup(self, modules):
        ip_tcl_files = []
        for module in modules:
            ip_tcl_files += module.get_ip_core_files()

        self._vivado_project = VivadoProject(
            name=self._project_name,
            modules=[],
            part=self._part_name,
            tcl_sources=ip_tcl_files
        )

        self._hash = self._calculate_hash(ip_tcl_files)

    @staticmethod
    def _calculate_hash(ip_files):
        """
        A string with hashes of the different IP core files.
        """
        data = ""

        def sort_by_file_name(path):
            return path.name

        for ip_file in sorted(ip_files, key=sort_by_file_name):
            with open(ip_file, "rb") as file_handle:
                ip_hash = hashlib.md5()
                ip_hash.update(file_handle.read())
                data += f"{ip_file}\n{ip_hash.hexdigest()}\n"

        return data

    def _save_hash(self):
        create_file(self._hash_file, self._hash)

    def _should_create(self):
        """
        Return True if a Vivado project create is needed, i.e. if anything has changed.
        """
        if not (self._hash_file.exists() and self.compile_order_file.exists()):
            return True

        return read_file(self._hash_file) != self._hash



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from tsfpga.module import BaseModule


class Module(BaseModule):

    def setup_simulations(self, vunit_proj, **kwargs):
        tb = vunit_proj.library(self.library_name).test_bench("tb_resync_pulse")
        for input_pulse_overload in [True, False]:
            name = "pulse_gating." if input_pulse_overload else ""

            generics = dict(input_pulse_overload=input_pulse_overload, output_clock_is_faster=True)
            tb.add_config(name=name + "output_clock_is_faster", generics=generics)

            generics = dict(input_pulse_overload=input_pulse_overload)
            tb.add_config(name=name + "output_clock_is_same", generics=generics)

            generics = dict(input_pulse_overload=input_pulse_overload, output_clock_is_slower=True)
            tb.add_config(name=name + "output_clock_is_slower", generics=generics)

        tb = vunit_proj.library(self.library_name).test_bench("tb_resync_counter")
        for pipeline_output in [True, False]:
            name = "pipeline_output" if pipeline_output else "dont_pipeline_output"

            generics = dict(pipeline_output=pipeline_output)
            tb.add_config(name=name, generics=generics)

        tb = vunit_proj.library(self.library_name).test_bench("tb_resync_cycles")
        for active_high in [True, False]:
            generics = dict(active_high=active_high, output_clock_is_faster=True)
            self.add_config(tb, generics=generics)

            generics = dict(active_high=active_high)
            self.add_config(tb, generics=generics)

            generics = dict(active_high=active_high, output_clock_is_slower=True)
            self.add_config(tb, generics=generics)



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

"""
Common functions and definitions in the example build environment.
"""

import tsfpga
from tsfpga.module import get_modules
from tsfpga.register_list import get_default_registers


TSFPGA_EXAMPLES_TEMP_DIR = tsfpga.TSFPGA_GENERATED


def get_tsfpga_modules(modules_folders=None):
    """
    Wrapper of the regular get_modules call with correct settings for tsfpga modules.
    """
    modules_folders = tsfpga.ALL_TSFPGA_MODULES_FOLDERS if modules_folders is None \
        else modules_folders
    return get_modules(modules_folders,
                       library_name_has_lib_suffix=False,
                       default_registers=get_default_registers())



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from pathlib import Path
import toml


def create_configuration(output_path,
                         modules=None,
                         vunit_proj=None,
                         vivado_location=None,
                         ip_core_vivado_project_sources_directory=None):
    """
    Create a configuration file (vhdl_ls.toml) for the rust_hdl VHDL Language Server.

    Can be used with modules and an "empty" VUnit project, or with a complete VUnit
    project with all user files added.

    Execution of this function takes roughly 12 ms for a large project (62 modules and a
    VUnit project).

    Args:
        output_path (`pathlib.Path`): Output folder.
        modules: A list of Module objects.
        vunit_proj: A VUnit project.
        vivado_location (`pathlib.Path`): Vivado binary path. Will add unisim from this Vivado installation.
        ip_core_vivado_project_sources_directory (`pathlib.Path`): Path to the sources directory of a
            Vivado project that contains generated "simulation" and "synthesis" files
            of IP cores (the "generate_target" TCL command). See simulate.py for an example
            of using this.
    """
    toml_data = dict(libraries=dict())

    if modules is not None:
        for module in modules:
            vhd_file_wildcard = module.path.resolve() / "**" / "*.vhd"
            toml_data["libraries"][module.library_name] = dict(
                files=[str(vhd_file_wildcard)])

    if vunit_proj is not None:
        for source_file in vunit_proj.get_compile_order():
            if source_file.library.name not in toml_data["libraries"]:
                toml_data["libraries"][source_file.library.name] = dict(
                    files=[])
            toml_data["libraries"][source_file.library.name]["files"].append(
                str(Path(source_file.name).resolve()))

    if vivado_location is not None:
        vcomponents_package = vivado_location.parent.parent / "data" / "vhdl" / "src" / "unisims" / "unisim_retarget_VCOMP.vhd"
        if not vcomponents_package.exists():
            raise FileNotFoundError(f"Could not find unisim file: {vcomponents_package}")

        toml_data["libraries"]["unisim"] = dict(files=[str(vcomponents_package.resolve())])

    if ip_core_vivado_project_sources_directory is not None:
        ip_sources_dir = ip_core_vivado_project_sources_directory / "ip"
        if not ip_sources_dir.exists():
            raise FileNotFoundError(f"Could not find IP sources dir: {ip_sources_dir}")

        # Add file from the "synth" folder rather than "sim". It seems that "synth"
        # always contains a VHDL file while "sim" sometimes contains a Verilog file.
        vhd_file_wildcard = ip_sources_dir.resolve() / "*" / "synth" / "*.vhd"
        toml_data["libraries"]["xil_defaultlib"] = dict(files=[str(vhd_file_wildcard)])

    with (output_path / "vhdl_ls.toml").open("w") as output_file_handle:
        toml.dump(toml_data, output_file_handle)



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from subprocess import CalledProcessError
import sys
import unittest

import pytest

import tsfpga
from tsfpga.constraint import Constraint
from tsfpga.module import get_modules
from tsfpga.system_utils import create_file, run_command
from tsfpga.vivado.project import VivadoProject
from tsfpga.test import file_contains_string


def test_building_artyz7_project(tmp_path):
    build_py = tsfpga.TSFPGA_EXAMPLES / "build.py"
    cmd = [
        sys.executable,
        build_py,
        "artyz7",
        "--project-path", tmp_path,
        "--output-path", tmp_path
    ]
    run_command(cmd, cwd=tsfpga.REPO_ROOT)
    assert (tmp_path / "artyz7.bit").exists()
    assert (tmp_path / "artyz7.bin").exists()
    assert (tmp_path / "artyz7.xsa").exists()


@pytest.mark.usefixtures("fixture_tmp_path")
class TestBasicProject(unittest.TestCase):

    tmp_path = None

    top_template = """
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library resync;


entity test_proj_top is
  port (
    clk_in : in std_logic;
    input : in std_logic;
    clk_out : in std_logic;
    output : out std_logic
  );
end entity;

architecture a of test_proj_top is
  signal input_p1 : std_logic;
begin

  pipe_input : process
  begin
    wait until rising_edge(clk_in);
    input_p1 <= input;
  end process;

{code_block}

end architecture;
"""

    def setUp(self):
        modules_folder = self.tmp_path / "modules"
        self.project_folder = self.tmp_path / "vivado"

        # Default top level
        resync = """
  code_block : entity resync.resync_level
  port map (
    data_in => input_p1,

    clk_out => clk_out,
    data_out => output
  );"""
        top = self.top_template.format(code_block=resync)
        self.top_file = create_file(modules_folder / "apa" / "test_proj_top.vhd", top)

        constraint = """
set_property -dict {package_pin H16 iostandard lvcmos33} [get_ports clk_in]
set_property -dict {package_pin P14 iostandard lvcmos33} [get_ports input]
set_property -dict {package_pin K17 iostandard lvcmos33} [get_ports clk_out]
set_property -dict {package_pin T16 iostandard lvcmos33} [get_ports output]

# 250 MHz
create_clock -period 4 -name clk_in [get_ports clk_in]
create_clock -period 4 -name clk_out [get_ports clk_out]
"""
        constraint_file = \
            create_file(modules_folder / "apa" / "test_proj_pinning.tcl", constraint)
        constraints = [Constraint(constraint_file)]

        modules = get_modules([modules_folder, tsfpga.TSFPGA_MODULES])
        self.proj = VivadoProject(
            name="test_proj",
            modules=modules,
            part="xc7z020clg400-1",
            constraints=constraints,
            # Faster
            default_run_index=2)
        self.proj.create(self.project_folder)

        self.log_file = self.project_folder / "vivado.log"

        self.runs_folder = self.project_folder / "test_proj.runs"

    def test_create_project(self):
        pass

    def test_synth_project(self):
        build_result = self.proj.build(self.project_folder, synth_only=True)
        assert build_result.success
        assert (self.runs_folder / "synth_2" / "hierarchical_utilization.rpt").exists()

    def test_synth_should_fail_if_source_code_does_not_compile(self):
        create_file(self.top_file, "garbage\napa\nhest")

        with pytest.raises(CalledProcessError):
            self.proj.build(self.project_folder, synth_only=True)
        assert file_contains_string(self.log_file, "\nERROR: Run synth_2 failed.")

    def test_synth_with_assert_false_should_fail(self):
        assert_false = """
  assert false severity failure;"""

        top = self.top_template.format(code_block=assert_false)
        create_file(self.top_file, top)

        with pytest.raises(CalledProcessError):
            self.proj.build(self.project_folder, synth_only=True, run_index=2)
        assert file_contains_string(self.log_file, "RTL assertion: \"Assertion violation.\"")

    def test_build_project(self):
        build_result = self.proj.build(self.project_folder, self.project_folder)

        assert (self.project_folder / (self.proj.name + ".bit")).exists()
        assert (self.project_folder / (self.proj.name + ".bin")).exists()
        assert (self.runs_folder / "impl_2" / "hierarchical_utilization.rpt").exists()

        # Sanity check some of the build result
        assert build_result.success

        assert build_result.synthesis_size["Total LUTs"] == 0, build_result.synthesis_size

        assert build_result.implementation_size["FFs"] > 0, build_result.implementation_size
        assert build_result.implementation_size["FFs"] < 2000, build_result.implementation_size

    def test_build_with_bad_timing_should_fail(self):
        # Do a ridiculously wide multiplication, which Vivado can't optimize away
        bad_timing = """
  mult_block : block
    signal resynced_input : std_logic;
  begin
    resync_level_inst : entity resync.resync_level
    port map (
      data_in => input_p1,

      clk_out => clk_out,
      data_out => resynced_input
    );

    mult : process
      constant bit_pattern : std_logic_vector(32 -1 downto 0) := x"deadbeef";
      variable term1, term2, term3 : unsigned(bit_pattern'range);
    begin
      wait until rising_edge(clk_out);
      term1 := unsigned(bit_pattern);
      term1(28) := resynced_input;

      term2 := unsigned(not bit_pattern);
      term2(22) := resynced_input;

      output <= xor (term1 * term2);
    end process;
  end block;"""

        top = self.top_template.format(code_block=bad_timing)
        create_file(self.top_file, top)

        with pytest.raises(CalledProcessError):
            self.proj.build(self.project_folder, self.project_folder)
        assert file_contains_string(self.log_file, "\nERROR: Timing not OK after implementation run.")

        assert (self.runs_folder / "impl_2" / "timing_summary.rpt").exists()

    def test_build_with_unhandled_clock_crossing_should_fail(self):
        bad_resync = """
  pipe_output : process
  begin
    wait until rising_edge(clk_out);
    output <= input_p1;
  end process;"""

        top = self.top_template.format(code_block=bad_resync)
        create_file(self.top_file, top)

        with pytest.raises(CalledProcessError):
            self.proj.build(self.project_folder, self.project_folder)
        assert file_contains_string(self.log_file, "\nERROR: Unhandled clock crossing in synth_2 run.")

        assert (self.runs_folder / "synth_2" / "hierarchical_utilization.rpt").exists()
        assert (self.runs_folder / "synth_2" / "timing_summary.rpt").exists()
        assert (self.runs_folder / "synth_2" / "clock_interaction.rpt").exists()



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------


class BuildStepTclHook:

    """
    Represent a TCL file that shall be used as hook in one of the build steps.
    """

    def __init__(self, tcl_file, hook_step):
        """
        Args:
            tcl_file (`pathlib.Path`): Path to a TCL file.
            hook_step (str): Name of a build step, e.g. "STEPS.ROUTE_DESIGN.TCL.PRE".
        """
        self.tcl_file = tcl_file
        self.hook_step = hook_step

    @property
    def step_is_synth(self):
        """
        True if the build step is in synthesis. False otherwise.
        """
        return "synth" in self.hook_step.lower()

    def __str__(self):
        result = str(self.__class__.__name__) + ":"
        return result + f"{self.hook_step}:{self.tcl_file}"



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------
# @eine has made a docker image that runs smoothly, run for example with:
#   docker run --rm --interactive --tty --volume ~/work/repo:/repo --workdir /repo/tsfpga ghdl/synth:formal /bin/bash
#   python3 examples/formal.py
# ------------------------------------------------------------------------------

import argparse
from pathlib import Path
import sys

PATH_TO_TSFPGA = Path(__file__).parent.parent.resolve()
sys.path.append(str(PATH_TO_TSFPGA))
PATH_TO_VUNIT = PATH_TO_TSFPGA.parent / "vunit"
sys.path.append(str(PATH_TO_VUNIT))

from tsfpga_example_env import get_tsfpga_modules, TSFPGA_EXAMPLES_TEMP_DIR
import tsfpga
from tsfpga.formal_project import FormalProject


def arguments(default_temp_dir=TSFPGA_EXAMPLES_TEMP_DIR):
    parser = argparse.ArgumentParser("Run formal tests",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--list-only",
                        action="store_true",
                        help="list the available tests")
    parser.add_argument("--project-path",
                        type=Path,
                        default=default_temp_dir / "formal_project",
                        help="the formal project will be placed here")
    args = parser.parse_args()

    return args


def main():
    args = arguments()
    modules = get_tsfpga_modules([tsfpga.TSFPGA_MODULES])
    formal_project = FormalProject(
        modules=modules,
        project_path=TSFPGA_EXAMPLES_TEMP_DIR / args.project_path)
    for module in modules:
        module.setup_formal(formal_project)

    if args.list_only:
        formal_project.list_tests()
        return

    formal_project.run()


if __name__ == "__main__":
    main()



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from tsfpga.register_types import Register


class RegisterCodeGenerator:

    """
    Common functions for generating register code.
    """

    @staticmethod
    def _iterate_registers(register_objects):
        for register_object in register_objects:
            if isinstance(register_object, Register):
                yield (register_object, None)
            else:
                for register in register_object.registers:
                    yield (register, register_object)



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

import fnmatch


class BuildProjectList:

    """
    Interface to handle a list of FPGA build projects.
    """

    def __init__(self, modules):
        """
        Args:
            modules (list(:class:`.BaseModule`)): Module objects that can define build projects.
        """
        self._modules = modules

    def get_projects(self, project_filters, include_netlist_not_top_builds=False):
        """
        Get projects matching any of the specified filters.

        Args:
            project_filters (list(str)): Project name filters.
            include_netlist_not_top_builds: Set True to get only netlist builds instead of only top level builds.

        Returns:
            A list of FPGA build project objects.
        """
        projects = list(self._iterate_projects(project_filters, include_netlist_not_top_builds))

        if not projects:
            raise ValueError(f"Could not find projects with filters: {project_filters}")

        return projects

    def list_projects(self, project_filters, include_netlist_not_top_builds=False):
        """
        Returns a string with a list of the projects matching the specified filters.

        Args:
            project_filters (list(str)): Project name filters.
            include_netlist_not_top_builds (bool): Set True to get only netlist builds (:class:`.VivadoNetlistProject`)
                instead of only top level builds (:class:`.VivadoProject`).

        Returns:
            A string listing projects matching the specified filters.
        """
        return "\n\n".join([str(project) for project in self._iterate_projects(project_filters, include_netlist_not_top_builds)])

    def _iterate_projects(self, project_filters=None, include_netlist_not_top_builds=False):
        projects = []
        for module in self._modules:
            projects += module.get_build_projects()
        for project in projects:
            if project.is_netlist_build == include_netlist_not_top_builds:
                if not project_filters:
                    yield project
                else:
                    for project_filter in project_filters:
                        if fnmatch.filter([project.name], project_filter):
                            yield project



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

import subprocess
import unittest

import pytest

import tsfpga
from tsfpga.system_utils import create_file, run_command
from tsfpga.register_list import from_toml


@pytest.mark.usefixtures("fixture_tmp_path")
class TestRegisterCompilation(unittest.TestCase):
    """
    Functional test: TOML -> registers -> Code generation -> compilation
    """

    tmp_path = None

    def setUp(self):
        self.working_dir = self.tmp_path
        self.include_dir = self.working_dir / "include"

        toml_file = tsfpga.TSFPGA_EXAMPLE_MODULES / "artyz7" / "regs_artyz7.toml"
        self.registers = from_toml("artyz7", toml_file)

        self.registers.add_constant("data_width", 24)
        self.registers.add_constant("decrement", -8)

    def test_compiling_c_header(self):
        main_file = self.working_dir / "main.c"
        main = """\
#include "artyz7_regs.h"
#include <assert.h>

int main()
{
  artyz7_regs_t regs;

  assert(sizeof(regs) == 4 * ARTYZ7_NUM_REGS);
  assert(ARTYZ7_DUMMY_REGS_CONFIGURATION_ADDR(3) == ARTYZ7_DUMMY_REGS_SETTINGS_ADDR(3) - 4);
  assert(ARTYZ7_DUMMY_REGS_CONFIGURATION_ADDR(0) == ARTYZ7_PLAIN_DUMMY_REG_ADDR + 4);
  assert(ARTYZ7_DUMMY_REGS_SETTINGS_ADDR(3) == 4 * (ARTYZ7_NUM_REGS - 1));

  assert(ARTYZ7_DUMMY_REGS_CONFIGURATION_ENABLE_BIT == 0);
  assert(ARTYZ7_DUMMY_REGS_CONFIGURATION_ENABLE == 1);
  assert(ARTYZ7_DUMMY_REGS_CONFIGURATION_DISABLE_BIT == 1);
  assert(ARTYZ7_DUMMY_REGS_CONFIGURATION_DISABLE == 2);

  regs.plain_dummy_reg = 0;
  regs.dummy_regs[0].configuration = ARTYZ7_DUMMY_REGS_CONFIGURATION_ENABLE;
  regs.dummy_regs[3].settings = (1 << ARTYZ7_DUMMY_REGS_CONFIGURATION_ENABLE_BIT);

  assert(ARTYZ7_CONSTANT_DATA_WIDTH == 24);
  assert(ARTYZ7_CONSTANT_DECREMENT == -8);

  return 0;
}
"""
        create_file(main_file, main)
        self.registers.create_c_header(self.include_dir)

        executable = self.working_dir / "artyz7.o"
        cmd = [
            "gcc",
            main_file,
            f"-o{executable}",
            f"-I{self.include_dir}"
        ]
        run_command(cmd)
        run_command([executable])

    def test_compiling_cpp(self):
        main_file = self.working_dir / "main.cpp"
        main = """\
#include <assert.h>

#include "include/artyz7.h"

int main()
{
  uint32_t data[fpga_regs::Artyz7::num_registers];
  volatile uint8_t *base_address = reinterpret_cast<volatile uint8_t *>(data);
  fpga_regs::Artyz7 artyz7 = fpga_regs::Artyz7(base_address);

  artyz7.set_plain_dummy_reg(1337);
  artyz7.set_dummy_regs_configuration(2, 0xfeed);
  artyz7.set_dummy_regs_settings(3, 0);
  artyz7.set_dummy_regs_configuration(3, fpga_regs::Artyz7::dummy_regs_configuration_enable);

  assert(artyz7.get_plain_dummy_reg() == 1337);
  assert(artyz7.get_dummy_regs_configuration(2) == 0xfeed);
  assert(artyz7.get_dummy_regs_settings(3) == 0);
  assert(artyz7.get_dummy_regs_configuration(3) == fpga_regs::Artyz7::dummy_regs_configuration_enable);

  assert(fpga_regs::Artyz7::dummy_regs_configuration_enable == 1);
  assert(fpga_regs::Artyz7::dummy_regs_configuration_disable == 2);

  assert(fpga_regs::Artyz7::data_width == 24);
  assert(fpga_regs::Artyz7::decrement == -8);

  return 0;
}
"""
        create_file(main_file, main)
        self.registers.create_cpp_interface(self.include_dir)
        self.registers.create_cpp_header(self.include_dir)
        self.registers.create_cpp_implementation(self.working_dir)
        cpp_class_file = self.working_dir / "artyz7.cpp"

        executable = self.working_dir / "artyz7.o"
        cmd = [
            "g++",
            main_file,
            cpp_class_file,
            f"-o{executable}",
            f"-I{self.include_dir}"
        ]
        run_command(cmd)
        run_command([executable])

    def test_setting_cpp_register_array_out_of_bounds_should_crash(self):
        main_file = self.working_dir / "main.cpp"
        main = """\
#include <assert.h>

#include "include/artyz7.h"

int main()
{
  uint32_t data[fpga_regs::Artyz7::num_registers];
  volatile uint8_t *base_address = reinterpret_cast<volatile uint8_t *>(data);
  fpga_regs::Artyz7 artyz7 = fpga_regs::Artyz7(base_address);

  artyz7.set_dummy_regs_settings(4, 1337);

  return 0;
}
"""
        create_file(main_file, main)
        self.registers.create_cpp_interface(self.include_dir)
        self.registers.create_cpp_header(self.include_dir)
        self.registers.create_cpp_implementation(self.working_dir)
        cpp_class_file = self.working_dir / "artyz7.cpp"

        executable = self.working_dir / "artyz7.o"
        cmd = [
            "g++",
            main_file,
            cpp_class_file,
            f"-o{executable}",
            f"-I{self.include_dir}"
        ]
        run_command(cmd)

        process = subprocess.Popen([executable], stderr=subprocess.PIPE)
        stderr = process.communicate()
        assert "Assertion `array_index < dummy_regs_array_length' failed" in str(stderr), stderr



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from pathlib import Path
from shutil import which
import subprocess


def run_vivado_tcl(vivado_path, tcl_file, no_log_file=False):
    """
    Setting cwd ensures that any .log or .jou files produced are placed in
    the same directory as the TCL file that produced them.

    Subprocess has to be run with shell=True on Windows where vivado is a bat file.

    Args:
        vivado_path (`pathlib.Path`): Path to Vivado executable. Can set to None
            to use whatever version is in PATH.
        tcl_file (`pathlib.Path`): Path to TCL file.
    """
    tcl_file = tcl_file.resolve()
    cmd = f"{get_vivado_path(vivado_path)} -mode batch -notrace -source {tcl_file}"
    if no_log_file:
        cmd += " -nojournal -nolog"
    subprocess.check_call(cmd, cwd=tcl_file.parent, shell=True)


def run_vivado_gui(vivado_path, project_file):
    """
    Setting cwd ensures that any .log or .jou files produced are placed in
    the same directory as the project.

    Subprocess has to be run with shell=True on Windows where vivado is a bat file.

    Args:
        vivado_path (`pathlib.Path`): Path to Vivado executable. Can set to None
            to use whatever version is in PATH.
        project_file (`pathlib.Path`): Path to a project .xpr file.
    """
    project_file = project_file.resolve()
    cmd = f"{get_vivado_path(vivado_path)} -mode gui {project_file}"
    if not project_file.exists():
        raise FileNotFoundError(f"Project does not exist: {project_file}")
    subprocess.check_call(cmd, cwd=project_file.parent, shell=True)


def get_vivado_path(vivado_path=None):
    """
    Wrapper to get a ``pathlib.Path`` to vivado executable.

    Args:
        vivado_path (`pathlib.Path`): Path to vivado executable. Set to None to use whatever
            is available in PATH.
    """
    if vivado_path is not None:
        return vivado_path.resolve()

    which_vivado = which("vivado")
    if which_vivado is None:
        raise FileNotFoundError("Could not find vivado on PATH")

    return Path(which_vivado).resolve()


def to_tcl_path(path):
    """
    Return a path string in a format suitable for TCL.
    """
    return str(path.resolve()).replace("\\", "/")



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from .simlib_commercial import VivadoSimlibCommercial
from .simlib_ghdl import VivadoSimlibGhdl


class VivadoSimlib:

    """
    Factory class for getting a Vivado simlib API.
    """

    @staticmethod
    def init(output_path, vunit_proj, vivado_path=None):
        """
        Get a Vivado simlib API suitable for your current simulator. Uses VUnit mechanism
        for detecting the simulator currently in use.

        Args:
            output_path (`pathlib.Path`): The compiled simlib will be placed here.
            vunit_proj: The VUnit project that is used to run simulation.
            vivado_path (`pathlib.Path`): Path to Vivado executable. If left out, the default
                from PATH will be used.
        Return:
            A :class:`.VivadoSimlibCommon` child object.
        """
        simulator_interface = vunit_proj._simulator_class  # pylint: disable=protected-access
        if simulator_interface.name == "ghdl":
            return VivadoSimlibGhdl(output_path, vunit_proj, simulator_interface, vivado_path)
        return VivadoSimlibCommercial(output_path, vunit_proj, simulator_interface, vivado_path)



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------


class SbyWriter:
    """
    Writes a Symbioyosys script file
    """

    # pylint: disable=too-many-arguments
    @staticmethod
    def write_sby(
            output_path,
            top,
            generics,
            formal_settings,
            compilation_outputs,
            src_files):
        """
        Create Symbiyosys script. This script will run ghdl --synth
        """

        p_args = [f"-P{compilation_output}" for compilation_output in compilation_outputs]
        g_args = []
        for key in iter(generics):
            g_args.append("-g" + str(key) + "=" + str(generics[key]))

        # Assume that the top name is the same as file name
        ghdl_synth_cmd = ["ghdl", "--std=08", "-fpsl"] + p_args + g_args + [top + ".vhd", "-e", top]

        sby = ""
        sby += "[options]\n"
        sby += f"mode {formal_settings['mode']} \n"
        sby += f"depth {str(formal_settings['depth'])} \n"
        sby += "\n"
        sby += "[engines]\n"
        sby += f"{formal_settings['engine_command']} \n"
        sby += "\n"
        sby += "[script]\n"
        sby += f"{' '.join(ghdl_synth_cmd)} \n"
        sby += f"prep -top {top} \n"
        sby += "\n"
        sby += "[files]\n"
        sby += "\n".join([str(src_file) for src_file in src_files])
        sby += "\n"

        with open(output_path, "w") as file_handle:
            file_handle.write(sby)



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------


def get_slogan():
    return "Tsfpga is a project platform for modern FPGA development."


def get_short_doc():
    return f"""{get_slogan()}
With its python build/simulation flow, along with complementary VHDL components, it is perfect \
for CI/CD and test-driven development.
Focus has been placed on flexibility and modularization, achieving scalability even in very \
large multi-vendor code bases.
"""


def get_doc():
    """
    Prepend get_short_doc() to this to get the complete doc.
    """
    return """Key features
------------

* Source code centric project structure: Build projects, test configurations, constraints, IP cores, etc. \
are handled "close" to the source code.
* Complete Vivado build system.
* Ideal for setting up VUnit simulation projects.
* Handling of IP cores and simlib for your simulation project, with automatic re-compile when necessary.
* Register code generation from TOML: VHDL package, HTML documentation, C header, C++ class.
* VHDL AXI components that enable the register bus: AXI-to-AXI-Lite converter, AXI-Lite interconnect, \
AXI-Lite mux (splitter), AXI-Lite clock domain crossing, AXI-Lite generic register file.

The maintainers place high focus on quality, with everything having good unit test coverage and a \
though-out structure.
The project is mature and used in many production environments.


License
-------

This project is released under the terms of the BSD 3-Clause License. See ``license.txt`` for details.\n"""



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

import re

import pytest

from tsfpga import TSFPGA_EXAMPLE_MODULES
from tsfpga.system_utils import create_file, read_file
from tsfpga.git_utils import find_git_files


class CopyrightHeader:

    def __init__(self, file, copyright_holder, copyright_text_lines=None):
        self._file = file
        self.comment_character = self._get_comment_character()
        self.separator_line = f"{self.comment_character} " + "-" * (79 - len(self.comment_character))
        self.expected_copyright_header = self._get_expected_copyright_header(
            copyright_holder, copyright_text_lines)

    def check_file(self):
        """
        Copyright comments should be correct. It should be followed by a blank line or another comment.
        """
        copyright_header_re = self.expected_copyright_header.replace("(", "\\(").replace(")", "\\)")
        regexp = re.compile(copyright_header_re + rf"($|\n|{self.comment_character})")
        data = read_file(self._file)
        return regexp.match(data) is not None

    def fix_file(self):
        if self._is_suitable_for_insertion():
            self._insert_copyright_header()
        else:
            raise ValueError(f"Can not fix copyright header in file {self._file}")

    def _get_expected_copyright_header(self, copyright_holder, copyright_text_lines):
        header = f"{self.separator_line}\n"
        header += f"{self.comment_character} Copyright (c) {copyright_holder}. All rights reserved.\n"
        if copyright_text_lines:
            header += f"{self.comment_character}\n"
            for copyright_text_line in copyright_text_lines:
                header += f"{self.comment_character} {copyright_text_line}\n"
        header += f"{self.separator_line}\n"
        return header

    def _get_comment_character(self):
        if self._file.name.endswith(".py"):
            return "#"
        if self._file.name.endswith(".vhd"):
            return "--"
        if self._file.name.endswith(".tcl"):
            return "#"
        if self._file.name.endswith(".c") or self._file.name.endswith(".cpp") or self._file.name.endswith(".h"):
            return "//"
        raise RuntimeError(f"Could not decide file ending of {self._file}")

    def _is_suitable_for_insertion(self):
        """
        If the file does not begin with a comment, we consired it suitable to insert a new copyright header comment.
        """
        return not read_file(self._file).startswith(self.comment_character)

    def _insert_copyright_header(self):
        data = read_file(self._file)
        data = f"{self.expected_copyright_header}\n{data}"
        create_file(self._file, data)


def files_to_check_for_copyright_header():
    file_endings = (".py", ".vhd", ".tcl")
    exclude_directories = [TSFPGA_EXAMPLE_MODULES / "artyz7" / "tcl"]
    for file_ending in file_endings:
        for file in find_git_files(file_endings_include=file_ending, exclude_directories=exclude_directories):
            yield file


def test_copyright_header_of_all_checked_in_files():
    test_ok = True
    for file in files_to_check_for_copyright_header():
        copyright_header_checker = CopyrightHeader(file, "Lukas Vik")
        if not copyright_header_checker.check_file():
            test_ok = False
            print(f"Fail for {file}.\nExpected:\n{copyright_header_checker.expected_copyright_header}")
    assert test_ok


def test_check_file(tmp_path):
    header = "-- " + "-" * 77 + "\n"
    header += "-- Copyright (c) Apa. All rights reserved.\n"
    header += "-- " + "-" * 77 + "\n"

    file = create_file(tmp_path / "header.vhd", header)
    copyright_header = CopyrightHeader(file, "Apa")
    assert copyright_header.check_file()

    file = create_file(tmp_path / "non_comment.vhd", header + "non-comment on line after")
    copyright_header = CopyrightHeader(file, "Apa")
    assert not copyright_header.check_file()

    file = create_file(tmp_path / "empty_line.vhd", header + "\nEmpty line and then non-comment")
    copyright_header = CopyrightHeader(file, "Apa")
    assert copyright_header.check_file()

    file = create_file(tmp_path / "further_comment.vhd", header + "-- Further comment\n")
    copyright_header = CopyrightHeader(file, "Apa")
    assert copyright_header.check_file()


def test_check_file_with_copyright_text(tmp_path):
    header = "-- " + "-" * 77 + "\n"
    header += "-- Copyright (c) Apa. All rights reserved.\n"
    header += "--\n"
    header += "-- Some more\n"
    header += "-- text.\n"
    header += "-- " + "-" * 77 + "\n"

    file = create_file(tmp_path / "header.vhd", header)
    copyright_header = CopyrightHeader(file, "Apa", ["Some more", "text."])
    assert copyright_header.check_file()


def test_fix_file_comment_insertion(tmp_path):
    data = "Apa\n"
    file = create_file(tmp_path / "file_for_test.vhd", data)

    copyright_header = CopyrightHeader(file, "Hest Hestsson")
    copyright_header.fix_file()

    data = read_file(file).split("\n")
    assert data[0] == "-- " + "-" * 77
    assert data[1] == "-- Copyright (c) Hest Hestsson. All rights reserved."
    assert data[2] == "-- " + "-" * 77
    assert data[3] == ""
    assert data[4] == "Apa"


def test_fix_file_should_not_run_on_dirty_file(tmp_path):
    data = "-- Custom comment line\n\nApa\n"
    file = create_file(tmp_path / "file_for_test.vhd", data)
    copyright_header = CopyrightHeader(file, "A")

    with pytest.raises(ValueError) as exception_info:
        copyright_header.fix_file()
    assert str(exception_info.value) == f"Can not fix copyright header in file {file}"



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from unittest import mock

from tsfpga.svn_utils import svn_local_changes_are_present

# SVN is pretty much impossible to work with. Most of the functions are untested.


@mock.patch("subprocess.check_output", autospec=True)
def test_svn_local_changes_are_present(check_output):
    check_output.return_value = """
?       .vscode
?       build
"""
    assert not svn_local_changes_are_present()

    check_output.return_value = """
?       .vscode
M       build.py
?       build
"""
    assert svn_local_changes_are_present()

    check_output.return_value = """
?       .vscode
!       build.py
?       build
"""
    assert svn_local_changes_are_present()
